;;; utf8.el -- Support for UTF8-encoded files.
;;; Author: Adrian Robert
;;; Based on "unicode.el" distributed by Otfried Cheong in the
;;; emacs-utf-0.62 distribution.

;; Copyright (C) 1998, 1999 Otfried Cheong

;; Version 0.5

;; Unicode.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published
;; by the Free Software Foundation; either version 2, or (at your
;; option) any later version.

;; This package is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(define-ccl-program ccl-encode-utf-8
  `(1
    ((read r0)
     (loop
      (write-read-repeat r0))))
  "CCL program to encode UTF-8 (does nothing!)")

(define-ccl-program ccl-decode-utf-8
  `(1
    ((read r0)
     (loop
      (write-read-repeat r0))))
  "CCL program to decode UTF-8 (does nothing!)")

(make-coding-system
 'unicode-utf8 4 ?U
 "Unicode UTF-8 encoding"
 '(ccl-decode-utf-8 . ccl-encode-utf-8)
 '((safe-charsets . t)
   (mime-charset . utf-8)
   (post-read-conversion . post-read-decode-utf8)
   (pre-write-conversion . pre-write-encode-utf8)))

(define-coding-system-alias 'utf8 'unicode-utf8)
(define-coding-system-alias 'utf-8 'unicode-utf8)

(defun post-read-decode-utf8 (len)
  (let ((pos (point))
	(buffer-modified-p (buffer-modified-p))
	(last-coding-system-used))
    (prog1
	(decode-utf8-region pos (+ pos len))
      (goto-char pos)
      (set-buffer-modified-p buffer-modified-p))))

(defun pre-write-encode-utf8 (from to)
  (let ((buf (current-buffer))
	(work (get-buffer-create " *pre-write-encoding-work*")))
    (set-buffer work)
    (erase-buffer)
    (if (stringp from)
	(insert from)
      (insert-buffer-substring buf from to))
    (let (last-coding-system-used)
      (encode-utf8-region 1 (point-max)))
    nil))

; ABR modified
(defun decode-utf8-region (beg end)
  "Decode a region from UTF-8.
Return the length of resulting text."
  (interactive "r")
  (let ((coding-system-for-read 'emacs-mule)
	(coding-system-for-write 'binary))
    (goto-char beg)
    (let* ((in-string (buffer-substring beg end))
           (out-string (ns-utf-to-mule in-string)))
      (delete-region beg end)
      (insert-string out-string)
;    (call-process-region beg end utf8-conversion-program t t nil)
      (- (point) beg))))

; ABR modified
(defun encode-utf8-region (beg end)
  "Encode a region to UTF-8."
  (let ((coding-system-for-write 'emacs-mule)
	(coding-system-for-read 'binary))
    (goto-char beg)
    (let* ((in-string (buffer-substring beg end))
           (out-string (ns-mule-to-utf in-string)))
      (delete-region beg end)
      (insert-string out-string))))

; Set up charsets
(define-charset nil 'korean-johab
  [2 96 2 0 ?= 0 "Johab" "Johab" "Johab Hangul elements"])

(define-charset nil 'unicode-mono
  [2 96 1 0 ?< 0 "Unicode Mono" "Unicode Mono" 
     "Unicode 0100-0dff,1000-10ff,13a0-18ff,1e00-28ff,fb00-ffff"])

(define-charset 151 'unicode-wide
  [2 96 2 0 ?> 0 "Unicode Wide" "Unicode Wide" 
     "Unicode 02e80-33ff, 4e00-9fa6 (characters not in JIS,KSC,GB,Big5)"])

;; Tell C code charset ID's of several charsets.
(setup-special-charsets)

;; use Unicode font for Ascii, Latin-1, Thai, Lao
(put-charset-property 'ascii 'x-charset-registry "iso10646")
(put-charset-property 'latin-iso8859-1 
		      'x-charset-registry "iso10646")
(put-charset-property 'unicode-mono 'x-charset-registry "iso10646")
(put-charset-property 'unicode-wide 'x-charset-registry "iso10646")
(put-charset-property 'korean-johab 'x-charset-registry "johab")
(put-charset-property 'lao	    'x-charset-registry "iso10646")
(put-charset-property 'thai-tis620  'x-charset-registry "iso10646")

(defvar mule-uni-1-table [
11904 11905 11906 11907 11908 11909 11910 11911 
11912 11913 11914 11915 11916 11917 11918 11919 
11920 11921 11922 11923 11924 11925 11926 11927 
11928 11929 11930 11931 11932 11933 11934 11935 
11936 11937 11938 11939 11940 11941 11942 11943 
11944 11945 11946 11947 11948 11949 11950 11951 
11952 11953 11954 11955 11956 11957 11958 11959 
11960 11961 11962 11963 11964 11965 11966 11967 
11968 11969 11970 11971 11972 11973 11974 11975 
11976 11977 11978 11979 11980 11981 11982 11983 
11984 11985 11986 11987 11988 11989 11990 11991 
11992 11993 11994 11995 11996 11997 11998 11999 
12000 12001 12002 12003 12004 12005 12006 12007 
12008 12009 12010 12011 12012 12013 12014 12015 
12016 12017 12018 12019 12020 12021 12022 12023 
12024 12025 12026 12027 12028 12029 12030 12031 
12032 12033 12034 12035 12036 12037 12038 12039 
12040 12041 12042 12043 12044 12045 12046 12047 
12048 12049 12050 12051 12052 12053 12054 12055 
12056 12057 12058 12059 12060 12061 12062 12063 
12064 12065 12066 12067 12068 12069 12070 12071 
12072 12073 12074 12075 12076 12077 12078 12079 
12080 12081 12082 12083 12084 12085 12086 12087 
12088 12089 12090 12091 12092 12093 12094 12095 
12096 12097 12098 12099 12100 12101 12102 12103 
12104 12105 12106 12107 12108 12109 12110 12111 
12112 12113 12114 12115 12116 12117 12118 12119 
12120 12121 12122 12123 12124 12125 12126 12127 
12128 12129 12130 12131 12132 12133 12134 12135 
12136 12137 12138 12139 12140 12141 12142 12143 
12144 12145 12146 12147 12148 12149 12150 12151 
12152 12153 12154 12155 12156 12157 12158 12159 
12160 12161 12162 12163 12164 12165 12166 12167 
12168 12169 12170 12171 12172 12173 12174 12175 
12176 12177 12178 12179 12180 12181 12182 12183 
12184 12185 12186 12187 12188 12189 12190 12191 
12192 12193 12194 12195 12196 12197 12198 12199 
12200 12201 12202 12203 12204 12205 12206 12207 
12208 12209 12210 12211 12212 12213 12214 12215 
12216 12217 12218 12219 12220 12221 12222 12223 
12224 12225 12226 12227 12228 12229 12230 12231 
12232 12233 12234 12235 12236 12237 12238 12239 
12240 12241 12242 12243 12244 12245 12246 12247 
12248 12249 12250 12251 12252 12253 12254 12255 
12256 12257 12258 12259 12260 12261 12262 12263 
12264 12265 12266 12267 12268 12269 12270 12271 
12272 12273 12274 12275 12276 12277 12278 12279 
12280 12281 12282 12283 12284 12285 12286 12287 
12292 12312 12313 12314 12315 12319 12320 12330 
12331 12332 12333 12334 12335 12336 12337 12338 
12339 12340 12341 12342 12343 12344 12345 12346 
12347 12348 12349 12350 12351 12352 12436 12437 
12438 12439 12440 12441 12442 12447 12448 12535 
12536 12537 12538 12543 12544 12545 12546 12547 
12548 12586 12587 12588 12589 12590 12591 12592 
12687 12688 12689 12690 12691 12692 12693 12694 
12695 12696 12697 12698 12699 12700 12701 12702 
12703 12704 12705 12706 12707 12708 12709 12710 
12711 12712 12713 12714 12715 12716 12717 12718 
12719 12720 12721 12722 12723 12724 12725 12726 
12727 12728 12729 12730 12731 12732 12733 12734 
12735 12736 12737 12738 12739 12740 12741 12742 
12743 12744 12745 12746 12747 12748 12749 12750 
12751 12752 12753 12754 12755 12756 12757 12758 
12759 12760 12761 12762 12763 12764 12765 12766 
12767 12768 12769 12770 12771 12772 12773 12774 
12775 12776 12777 12778 12779 12780 12781 12782 
12783 12784 12785 12786 12787 12788 12789 12790 
12791 12792 12793 12794 12795 12796 12797 12798 
12799 12829 12830 12831 12842 12843 12844 12845 
12846 12847 12848 12849 12850 12851 12852 12853 
12854 12855 12856 12857 12858 12859 12860 12861 
12862 12863 12864 12865 12866 12867 12868 12869 
12870 12871 12872 12873 12874 12875 12876 12877 
12878 12879 12880 12881 12882 12883 12884 12885 
12886 12887 12888 12889 12890 12891 12892 12893 
12894 12895 12924 12925 12926 12928 12929 12930 
12931 12932 12933 12934 12935 12936 12937 12938 
12939 12940 12941 12942 12943 12944 12945 12946 
12947 12948 12949 12950 12951 12952 12953 12954 
12955 12956 12957 12958 12959 12960 12961 12962 
12964 12965 12966 12967 12968 12969 12970 12971 
12972 12973 12974 12975 12976 12977 12978 12979 
12980 12981 12982 12983 12984 12985 12986 12987 
12988 12989 12990 12991 12992 12993 12994 12995 
12996 12997 12998 12999 13000 13001 13002 13003 
13004 13005 13006 13007 13008 13009 13010 13011 
13012 13013 13014 13015 13016 13017 13018 13019 
13020 13021 13022 13023 13024 13025 13026 13027 
13028 13029 13030 13031 13032 13033 13034 13035 
13036 13037 13038 13039 13040 13041 13042 13043 
13044 13045 13046 13047 13048 13049 13050 13051 
13052 13053 13054 13055 13056 13057 13058 13059 
13060 13061 13062 13063 13064 13065 13066 13067 
13068 13069 13070 13071 13072 13073 13074 13075 
13076 13077 13078 13079 13080 13081 13082 13083 
13084 13085 13086 13087 13088 13089 13090 13091 
13092 13093 13094 13095 13096 13097 13098 13099 
13100 13101 13102 13103 13104 13105 13106 13107 
13108 13109 13110 13111 13112 13113 13114 13115 
13116 13117 13118 13119 13120 13121 13122 13123 
13124 13125 13126 13127 13128 13129 13130 13131 
13132 13133 13134 13135 13136 13137 13138 13139 
13140 13141 13142 13143 13144 13145 13146 13147 
13148 13149 13150 13151 13152 13153 13154 13155 
13156 13157 13158 13159 13160 13161 13162 13163 
13164 13165 13166 13167 13168 13169 13170 13171 
13172 13173 13174 13175 13176 13177 13178 13179 
13180 13181 13182 13183 13189 13190 13191 13259 
13260 13261 13268 13271 13273 13274 13278 13279 
13280 13281 13282 13283 13284 13285 13286 13287 
13288 13289 13290 13291 13292 13293 13294 13295 
13296 13297 13298 13299 13300 13301 13302 13303 
13304 13305 13306 13307 13308 13309 13310 13311 
19970 19972 19973 19974 19986 20000 20003 20009 
20015 20021 20023 20032 20033 20036 20038 20042 
20049 20058 20059 20067 20068 20069 20071 20072 
20074 20076 20078 20079 20082 20084 20085 20088 
20089 20090 20091 20092 20093 20097 20103 20118 
20119 20125 20131 20138 20143 20145 20148 20151 
20152 20156 20157 20158 20172 20176 20178 20187 
20192 20194 20198 20199 20207 20216 20217 20218 
20220 20222 20227 20230 20231 20236 20246 20257 
20259 20264 20265 20270 20273 20277 20279 20281 
20288 20290 20292 20293 20298 20299 20326 20328 
20333 20337 20338 20364 20366 20371 20377 20383 
20386 20388 20400 20401 20404 20408 20412 20413 
20414 20422 20424 20428 20434 20437 20450 20455 
20459 20464 20466 20473 20475 20476 20477 20479 
20481 20483 20484 20488 20490 20496 20509 20530 
20532 20536 20541 20543 20546 20548 20562 20564 
20568 20569 20582 20583 20593 20601 20604 20609 
20612 20614 20617 20618 20623 20624 20627 20631 
20639 20640 20641 20644 20646 20650 20651 20665 
20668 20672 20675 20684 20688 20690 20696 20697 
20700 20703 20705 20706 20715 20722 20724 20727 
20730 20732 20749 20750 20751 20761 20763 20765 
20766 20771 20775 20776 20779 20780 20783 20790 
20798 20802 20810 20815 20819 20824 20832 20836 
20838 20847 20850 20858 20862 20863 20867 20868 
20875 20878 20890 20893 20895 20897 20899 20903 
20909 20910 20920 20922 20927 20929 20930 20931 
20946 20947 20949 20953 20954 20959 20962 20963 
20965 20968 20972 20974 20978 20980 20983 20997 
21007 21013 21023 21026 21027 21030 21036 21039 
21044 21052 21054 21061 21064 21075 21079 21080 
21081 21088 21094 21096 21110 21113 21118 21125 
21126 21134 21135 21141 21146 21148 21157 21159 
21167 21168 21174 21175 21176 21178 21181 21188 
21190 21192 21194 21196 21198 21199 21201 21204 
21210 21212 21217 21221 21224 21226 21228 21229 
21230 21238 21245 21251 21252 21255 21260 21267 
21268 21272 21275 21278 21284 21285 21287 21288 
21289 21291 21292 21298 21301 21302 21314 21323 
21328 21337 21339 21341 21349 21354 21357 21366 
21374 21376 21377 21379 21383 21384 21392 21393 
21403 21409 21418 21419 21423 21425 21429 21431 
21432 21434 21436 21437 21438 21440 21444 21445 
21446 21447 21455 21456 21458 21459 21461 21466 
21468 21469 21470 21479 21492 21502 21503 21504 
21506 21509 21524 21530 21538 21562 21567 21572 
21580 21581 21594 21597 21598 21609 21613 21614 
21625 21635 21637 21641 21642 21651 21655 21660 
21662 21663 21685 21689 21706 21707 21723 21731 
21740 21743 21744 21748 21749 21750 21753 21758 
21760 21762 21773 21779 21781 21784 21785 21789 
21790 21791 21793 21797 21800 21801 21803 21818 
21821 21826 21831 21844 21848 21849 21850 21851 
21856 21864 21865 21867 21871 21872 21873 21874 
21875 21876 21881 21882 21893 21904 21910 21911 
21915 21920 21935 21940 21944 21946 21948 21953 
21975 21976 21982 21984 21997 21998 22000 22001 
22004 22008 22011 22019 22021 22023 22026 22027 
22033 22041 22042 22048 22049 22050 22053 22054 
22056 22059 22071 22076 22083 22084 22087 22091 
22095 22097 22098 22101 22102 22109 22111 22113 
22119 22133 22141 22152 22153 22155 22161 22162 
22166 22171 22174 22175 22177 22180 22185 22192 
22193 22200 22201 22202 22203 22207 22212 22215 
22223 22224 22226 22229 22230 22233 22236 22246 
22248 22249 22252 22255 22262 22264 22267 22268 
22277 22286 22288 22289 22293 22295 22297 22301 
22305 22308 22309 22315 22321 22322 22325 22326 
22332 22333 22335 22339 22340 22342 22344 22355 
22356 22357 22358 22360 22361 22371 22373 22375 
22380 22392 22393 22394 22398 22401 22407 22410 
22413 22414 22416 22417 22422 22428 22440 22444 
22447 22449 22455 22459 22462 22468 22469 22473 
22474 22477 22481 22487 22491 22494 22502 22504 
22506 22507 22526 22527 22531 22543 22546 22547 
22550 22551 22552 22554 22559 22562 22566 22571 
22588 22590 22594 22595 22597 22598 22599 22608 
22614 22620 22623 22624 22630 22631 22634 22638 
22643 22647 22648 22660 22668 22669 22677 22678 
22683 22690 22692 22695 22698 22701 22704 22706 
22708 22709 22710 22711 22720 22723 22724 22731 
22733 22736 22752 22753 22758 22762 22765 22773 
22776 22784 22785 22789 22792 22795 22801 22803 
22813 22814 22817 22819 22822 22824 22832 22835 
22837 22838 22843 22845 22847 22851 22854 22860 
22861 22866 22873 22875 22877 22878 22879 22883 
22884 22886 22895 22901 22906 22921 22923 22924 
22929 22932 22933 22938 22940 22943 22957 22960 
22967 22968 22975 22978 22980 22997 23007 23010 
23015 23023 23024 23042 23051 23053 23054 23056 
23058 23060 23069 23073 23074 23076 23078 23079 
23080 23082 23083 23084 23088 23098 23099 23101 
23103 23109 23115 23118 23119 23124 23129 23137 
23139 23144 23147 23150 23151 23153 23154 23155 
23161 23166 23168 23169 23170 23173 23174 23175 
23176 23177 23181 23185 23190 23192 23193 23200 
23201 23203 23208 23211 23213 23235 23237 23246 
23247 23249 23251 23268 23271 23279 23280 23282 
23292 23294 23296 23300 23302 23306 23309 23310 
23313 23314 23317 23320 23324 23327 23337 23339 
23345 23347 23349 23353 23354 23355 23361 23362 
23364 23366 23369 23370 23375 23378 23390 23392 
23393 23399 23400 23405 23407 23412 23414 23417 
23422 23426 23430 23434 23440 23441 23444 23465 
23471 23473 23474 23479 23482 23483 23484 23496 
23503 23509 23514 23516 23533 23540 23543 23549 
23552 23575 23579 23582 23585 23587 23590 23593 
23595 23598 23599 23602 23605 23606 23619 23634 
23639 23642 23647 23659 23664 23666 23669 23671 
23672 23677 23680 23683 23684 23685 23687 23694 
23710 23730 23732 23737 23746 23747 23757 23761 
23765 23772 23773 23778 23779 23783 23787 23794 
23795 23797 23804 23806 23812 23813 23816 23817 
23818 23824 23827 23836 23841 23850 23851 23852 
23855 23867 23876 23878 23880 23885 23887 23891 
23892 23894 23895 23898 23903 23904 23905 23908 
23910 23914 23917 23918 23920 23925 23928 23931 
23939 23941 23950 23951 23953 23958 23960 23963 
23971 23972 23973 23974 23979 23987 23990 23993 
23995 23998 23999 24001 24004 24008 24010 24014 
24016 24023 24025 24026 24028 24036 24042 24044 
24045 24054 24056 24058 24060 24064 24071 24073 
24077 24082 24083 24094 24106 24108 24114 24117 
24121 24122 24134 24136 24137 24144 24145 24146 
24150 24154 24165 24177 24197 24206 24210 24216 
24221 24225 24233 24239 24250 24251 24252 24253 
24255 24256 24269 24292 24299 24301 24309 24313 
24317 24326 24334 24345 24348 24350 24353 24363 
24364 24370 24372 24379 24381 24383 24386 24389 
24391 24397 24410 24414 24415 24416 24419 24424 
24430 24434 24437 24442 24461 24462 24463 24468 
24474 24477 24482 24483 24484 24487 24496 24497 
24504 24506 24514 24519 24520 24522 24523 24526 
24531 24538 24539 24543 24546 24550 24553 24556 
24562 24566 24569 24572 24583 24584 24600 24607 
24611 24624 24630 24632 24637 24638 24648 24654 
24655 24657 24662 24663 24668 24673 24689 24692 
24702 24706 24719 24721 24723 24728 24729 24734 
24737 24741 24750 24770 24784 24786 24790 24791 
24798 24805 24810 24829 24834 24839 24844 24849 
24855 24857 24862 24866 24869 24874 24877 24881 
24883 24885 24888 24889 24890 24899 24912 24919 
24924 24928 24937 24941 24952 24955 24959 24964 
24965 24966 24968 24975 24981 24983 24988 24990 
24992 24995 24997 24998 25017 25019 25021 25024 
25028 25038 25039 25043 25045 25047 25049 25050 
25051 25052 25053 25057 25058 25068 25071 25075 
25090 25093 25107 25116 25128 25141 25145 25148 
25156 25157 25167 25174 25175 25181 25183 25205 
25208 25221 25227 25229 25232 25241 25245 25251 
25254 25255 25266 25271 25274 25280 25281 25283 
25301 25310 25316 25322 25348 25349 25350 25354 
25362 25367 25368 25372 25382 25390 25392 25393 
25397 25399 25426 25427 25435 25437 25440 25444 
25446 25450 25452 25459 25460 25465 25470 25478 
25483 25491 25493 25498 25510 25526 25529 25535 
25537 25553 25556 25570 25574 25580 25595 25596 
25598 25603 25604 25607 25608 25617 25625 25629 
25641 25649 25650 25656 25659 25660 25673 25676 
25679 25685 25686 25687 25690 25698 25699 25700 
25706 25713 25724 25726 25728 25729 25734 25741 
25742 25748 25755 25761 25767 25768 25770 25775 
25780 25782 25798 25800 25804 25809 25811 25813 
25820 25821 25823 25834 25838 25845 25849 25858 
25864 25866 25867 25873 25882 25886 25887 25895 
25896 25904 25905 25914 25916 25920 25922 25924 
25927 25931 25936 25938 25946 25951 25952 25953 
25961 25966 25969 25981 25990 25994 25997 25999 
26008 26010 26019 26022 26033 26036 26037 26042 
26046 26048 26055 26056 26057 26058 26065 26068 
26069 26072 26076 26084 26090 26091 26104 26105 
26110 26113 26134 26135 26136 26138 26139 26147 
26153 26154 26156 26167 26168 26173 26176 26182 
26184 26189 26190 26192 26200 26208 26211 26221 
26229 26237 26239 26245 26255 26258 26259 26266 
26267 26268 26270 26275 26276 26277 26284 26285 
26294 26303 26306 26307 26309 26317 26318 26320 
26321 26323 26324 26325 26327 26335 26337 26338 
26341 26343 26346 26351 26353 26363 26370 26374 
26375 26380 26382 26385 26393 26394 26396 26404 
26405 26409 26416 26418 26436 26442 26450 26452 
26456 26459 26470 26471 26475 26478 26496 26498 
26504 26506 26518 26523 26532 26540 26545 26556 
26557 26559 26567 26581 26582 26583 26593 26600 
26617 26625 26630 26637 26640 26645 26649 26651 
26659 26660 26663 26668 26672 26678 26687 26695 
26706 26710 26711 26712 26714 26715 26718 26730 
26732 26736 26739 26746 26756 26760 26766 26773 
26776 26777 26778 26782 26789 26807 26808 26811 
26813 26814 26815 26817 26819 26831 26841 26843 
26850 26853 26861 26878 26879 26882 26883 26889 
26902 26904 26909 26910 26921 26923 26924 26926 
26929 26938 26942 26947 26950 26951 26957 26960 
26980 26983 26994 27007 27013 27019 27020 27023 
27026 27037 27039 27042 27064 27066 27072 27077 
27080 27089 27090 27093 27094 27095 27098 27100 
27105 27107 27113 27119 27120 27125 27130 27139 
27150 27152 27162 27164 27172 27180 27181 27184 
27187 27191 27202 27203 27205 27210 27212 27223 
27228 27235 27244 27246 27248 27251 27252 27253 
27255 27258 27259 27261 27266 27270 27272 27274 
27275 27279 27288 27289 27293 27303 27312 27313 
27314 27317 27324 27326 27327 27328 27332 27336 
27337 27338 27342 27346 27348 27349 27350 27351 
27352 27362 27363 27364 27366 27369 27373 27378 
27380 27381 27383 27389 27390 27391 27393 27398 
27399 27405 27406 27412 27413 27419 27420 27430 
27433 27434 27435 27438 27440 27445 27456 27460 
27471 27474 27479 27480 27482 27485 27496 27499 
27500 27502 27504 27505 27509 27514 27517 27525 
27536 27548 27549 27560 27561 27564 27576 27577 
27582 27601 27612 27613 27621 27625 27629 27630 
27633 27636 27638 27642 27658 27662 27666 27676 
27678 27693 27697 27705 27706 27708 27716 27717 
27720 27729 27731 27734 27736 27747 27756 27758 
27765 27767 27772 27775 27793 27799 27806 27808 
27816 27823 27829 27848 27851 27854 27864 27871 
27876 27878 27892 27903 27906 27909 27910 27923 
27924 27925 27932 27937 27939 27940 27942 27977 
27980 27984 27989 27990 27991 27995 27997 28011 
28017 28018 28019 28033 28047 28054 28058 28066 
28069 28077 28080 28081 28086 28089 28097 28110 
28135 28158 28159 28161 28162 28164 28166 28175 
28178 28184 28190 28202 28215 28226 28232 28236 
28239 28240 28247 28249 28266 28268 28269 28272 
28277 28283 28284 28285 28289 28292 28295 28298 
28299 28305 28309 28314 28328 28329 28332 28333 
28341 28344 28377 28379 28387 28391 28394 28400 
28403 28410 28420 28427 28428 28432 28438 28439 
28443 28445 28456 28468 28477 28482 28484 28488 
28489 28490 28492 28502 28505 28517 28520 28529 
28533 28537 28547 28554 28559 28568 28569 28570 
28571 28573 28575 28597 28599 28603 28613 28624 
28627 28630 28631 28633 28645 28650 28661 28664 
28665 28669 28674 28675 28680 28688 28690 28691 
28709 28717 28718 28726 28733 28743 28747 28749 
28750 28755 28756 28761 28764 28775 28782 28786 
28787 28791 28793 28795 28801 28807 28808 28811 
28812 28813 28815 28816 28823 28827 28830 28832 
28834 28835 28837 28838 28839 28840 28842 28850 
28854 28863 28868 28873 28876 28880 28885 28886 
28899 28901 28906 28910 28914 28917 28926 28929 
28931 28933 28935 28936 28943 28945 28946 28957 
28964 28967 28969 28970 28971 28972 28973 28979 
28980 28981 28983 28984 28985 28987 28989 28990 
28991 28992 29000 29007 29009 29015 29019 29035 
29037 29039 29041 29044 29045 29046 29047 29049 
29052 29054 29055 29059 29067 29068 29069 29070 
29073 29075 29078 29091 29094 29099 29101 29102 
29108 29110 29111 29115 29132 29133 29137 29139 
29149 29150 29155 29161 29162 29163 29167 29171 
29174 29175 29178 29188 29192 29193 29195 29198 
29199 29201 29202 29205 29206 29207 29208 29212 
29216 29217 29220 29221 29227 29230 29231 29235 
29236 29251 29253 29262 29264 29265 29268 29269 
29271 29276 29284 29285 29288 29291 29293 29297 
29315 29332 29337 29340 29355 29361 29363 29371 
29372 29374 29383 29391 29395 29397 29405 29413 
29415 29429 29442 29444 29445 29446 29449 29453 
29456 29460 29466 29472 29476 29480 29501 29505 
29510 29511 29512 29515 29523 29524 29525 29526 
29532 29540 29553 29556 29561 29580 29581 29583 
29584 29593 29594 29596 29598 29603 29610 29617 
29629 29633 29636 29653 29663 29665 29666 29676 
29679 29680 29683 29691 29698 29710 29713 29714 
29716 29717 29719 29720 29721 29724 29726 29727 
29735 29751 29753 29755 29757 29758 29765 29767 
29768 29769 29772 29779 29782 29784 29789 29793 
29798 29803 29804 29812 29816 29818 29819 29828 
29836 29837 29839 29841 29843 29846 29849 29851 
29853 29860 29868 29870 29875 29876 29884 29892 
29895 29896 29897 29900 29901 29902 29904 29905 
29907 29930 29931 29933 29939 29945 29946 29948 
29958 29961 29962 29984 29988 29991 30004 30006 
30017 30018 30019 30032 30034 30037 30038 30039 
30040 30046 30049 30056 30062 30065 30075 30076 
30081 30085 30088 30093 30099 30107 30108 30110 
30118 30120 30121 30125 30135 30150 30163 30172 
30181 30188 30190 30210 30212 30214 30215 30222 
30226 30252 30254 30262 30263 30265 30273 30276 
30277 30282 30283 30287 30289 30293 30299 30301 
30310 30323 30324 30327 30341 30348 30349 30356 
30359 30360 30367 30368 30369 30370 30371 30373 
30375 30376 30377 30380 30387 30390 30396 30400 
30401 30407 30411 30412 30421 30425 30432 30434 
30440 30443 30454 30461 30463 30464 30466 30470 
30478 30479 30484 30486 30487 30488 30492 30497 
30506 30507 30508 30510 30512 30527 30530 30536 
30537 30547 30551 30552 30557 30564 30576 30577 
30578 30579 30580 30581 30582 30583 30584 30586 
30587 30598 30602 30608 30611 30612 30614 30616 
30628 30630 30638 30639 30648 30654 30656 30657 
30659 30661 30662 30664 30667 30673 30674 30678 
30685 30687 30689 30692 30694 30698 30708 30709 
30724 30727 30728 30730 30731 30745 30747 30750 
30756 30774 30781 30785 30786 30788 30790 30795 
30801 30803 30804 30808 30809 30810 30811 30815 
30817 30819 30822 30823 30835 30837 30838 30840 
30842 30845 30850 30856 30858 30859 30864 30866 
30877 30886 30894 30902 30903 30904 30909 30911 
30912 30914 30919 30930 30931 30934 30935 30936 
30940 30948 30950 30955 30958 30960 30961 30966 
30968 30976 30979 30982 30984 30986 30987 30989 
30991 30997 31000 31002 31007 31008 31010 31022 
31024 31026 31027 31030 31031 31043 31053 31054 
31064 31065 31078 31084 31086 31089 31093 31094 
31099 31102 31107 31110 31111 31113 31116 31129 
31134 31135 31139 31141 31145 31151 31157 31164 
31170 31172 31175 31180 31184 31187 31188 31191 
31193 31194 31195 31202 31205 31208 31217 31218 
31219 31220 31221 31225 31228 31230 31231 31233 
31239 31241 31247 31254 31261 31265 31268 31269 
31271 31273 31274 31276 31277 31284 31285 31288 
31290 31297 31301 31314 31315 31317 31321 31325 
31326 31332 31333 31334 31338 31343 31346 31347 
31356 31362 31374 31386 31387 31393 31396 31399 
31417 31419 31420 31421 31426 31430 31433 31436 
31438 31440 31444 31451 31463 31464 31465 31468 
31473 31474 31475 31476 31477 31484 31486 31495 
31500 31501 31510 31511 31516 31519 31521 31527 
31529 31543 31549 31551 31553 31554 31555 31571 
31573 31575 31577 31580 31582 31583 31592 31594 
31595 31599 31612 31615 31617 31619 31625 31635 
31646 31651 31653 31662 31664 31666 31667 31670 
31674 31675 31676 31677 31679 31682 31683 31685 
31688 31693 31696 31702 31703 31724 31727 31733 
31738 31748 31752 31765 31768 31770 31771 31780 
31790 31791 31793 31794 31796 31797 31798 31802 
31810 31812 31814 31819 31822 31825 31826 31829 
31837 31838 31841 31842 31848 31853 31856 31857 
31862 31863 31878 31879 31886 31887 31891 31897 
31898 31904 31910 31913 31916 31920 31926 31927 
31928 31939 31940 31942 31945 31951 31955 31962 
31963 31969 31971 31972 31973 31974 31977 31979 
31981 31987 31989 31993 31996 31999 32035 32036 
32037 32038 32042 32045 32052 32055 32072 32073 
32077 32087 32089 32090 32093 32096 32100 32101 
32108 32116 32120 32126 32130 32135 32138 32139 
32144 32149 32151 32152 32164 32165 32168 32179 
32182 32195 32200 32205 32208 32211 32212 32226 
32229 32235 32237 32245 32248 32252 32253 32254 
32255 32256 32258 32262 32263 32280 32281 32295 
32296 32300 32334 32335 32347 32356 32357 32364 
32366 32369 32388 32389 32413 32414 32430 32436 
32443 32444 32470 32484 32492 32505 32522 32528 
32542 32567 32571 32572 32576 32577 32582 32583 
32585 32594 32595 32598 32601 32612 32623 32640 
32644 32655 32656 32659 32663 32664 32665 32682 
32683 32692 32708 32712 32723 32726 32729 32730 
32733 32740 32743 32758 32762 32770 32776 32778 
32787 32794 32797 32803 32811 32814 32815 32818 
32826 32828 32832 32833 32836 32837 32841 32846 
32851 32853 32855 32857 32859 32864 32867 32869 
32870 32875 32877 32878 32890 32891 32892 32897 
32904 32909 32910 32913 32916 32919 32926 32934 
32935 32936 32944 32947 32950 32953 32955 32971 
32978 32979 32991 32994 33006 33015 33023 33027 
33028 33035 33036 33047 33052 33056 33062 33064 
33066 33070 33076 33077 33079 33083 33084 33087 
33088 33089 33090 33093 33097 33110 33111 33112 
33117 33123 33128 33130 33132 33141 33153 33156 
33157 33166 33168 33170 33172 33174 33185 33189 
33197 33199 33206 33227 33230 33236 33238 33244 
33252 33259 33263 33264 33265 33269 33270 33277 
33283 33295 33299 33305 33306 33316 33318 33319 
33345 33347 33350 33352 33354 33356 33357 33364 
33376 33381 33383 33403 33409 33414 33415 33417 
33420 33429 33430 33458 33478 33481 33488 33498 
33501 33506 33513 33518 33528 33532 33535 33546 
33547 33552 33554 33555 33565 33567 33569 33582 
33584 33597 33598 33621 33623 33625 33629 33634 
33648 33649 33650 33657 33664 33666 33668 33681 
33697 33708 33709 33723 33726 33741 33744 33746 
33754 33766 33773 33792 33794 33797 33800 33812 
33813 33814 33815 33817 33818 33822 33823 33825 
33838 33854 33857 33864 33866 33871 33875 33877 
33880 33892 33898 33906 33915 33916 33919 33920 
33921 33923 33925 33930 33938 33939 33941 33942 
33955 33957 33958 33971 33973 33975 33982 33987 
33992 34005 34008 34012 34014 34017 34018 34020 
34029 34037 34040 34049 34051 34052 34053 34064 
34075 34082 34098 34099 34100 34102 34114 34124 
34127 34128 34130 34140 34143 34159 34160 34163 
34173 34175 34194 34195 34213 34221 34226 34235 
34236 34250 34252 34260 34262 34267 34272 34279 
34286 34291 34292 34293 34300 34307 34312 34317 
34318 34319 34320 34322 34324 34325 34333 34344 
34347 34359 34365 34369 34370 34372 34373 34377 
34378 34391 34392 34397 34400 34406 34418 34421 
34422 34424 34435 34436 34440 34441 34447 34450 
34459 34463 34464 34470 34477 34478 34482 34514 
34517 34528 34529 34533 34535 34556 34557 34559 
34575 34576 34580 34582 34589 34591 34603 34607 
34614 34617 34621 34628 34629 34634 34673 34674 
34688 34694 34698 34700 34702 34709 34713 34720 
34721 34725 34726 34727 34729 34737 34753 34765 
34766 34767 34773 34774 34778 34793 34800 34801 
34805 34808 34813 34820 34823 34834 34840 34842 
34846 34861 34868 34874 34887 34889 34895 34896 
34897 34904 34908 34911 34912 34918 34931 34936 
34938 34939 34950 34951 34954 34959 34960 34964 
34973 34976 34979 34981 34982 34985 34991 34995 
35003 35015 35016 35025 35027 35040 35046 35049 
35050 35053 35071 35072 35075 35080 35085 35087 
35100 35104 35108 35112 35129 35130 35135 35136 
35141 35143 35144 35146 35150 35156 35157 35173 
35175 35176 35184 35189 35192 35197 35200 35204 
35209 35212 35213 35214 35216 35217 35218 35220 
35225 35232 35237 35240 35243 35248 35249 35252 
35253 35256 35259 35260 35267 35277 35287 35288 
35306 35310 35317 35321 35325 35329 35333 35334 
35337 35339 35341 35348 35353 35354 35356 35360 
35361 35364 35366 35368 35369 35371 35374 35375 
35378 35381 35383 35384 35389 35394 35395 35399 
35401 35403 35411 35418 35420 35421 35423 35428 
35429 35431 35434 35439 35448 35453 35454 35456 
35464 35470 35472 35476 35479 35483 35484 35485 
35487 35490 35497 35502 35503 35505 35507 35508 
35509 35511 35521 35530 35534 35536 35555 35557 
35561 35564 35577 35581 35587 35593 35615 35625 
35629 35634 35636 35640 35647 35651 35652 35661 
35678 35681 35682 35684 35689 35694 35699 35701 
35702 35708 35713 35719 35721 35725 35727 35729 
35735 35739 35741 35756 35761 35771 35783 35792 
35818 35849 35870 35896 35904 35908 35921 35922 
35923 35928 35929 35931 35934 35936 35939 35943 
35950 35956 35966 35967 35971 35975 35976 35979 
35990 35995 35999 36006 36017 36038 36041 36043 
36048 36052 36054 36056 36073 36075 36079 36082 
36086 36087 36095 36097 36099 36108 36110 36113 
36114 36120 36122 36124 36128 36177 36178 36183 
36191 36197 36202 36218 36220 36222 36223 36226 
36227 36230 36231 36232 36247 36248 36250 36253 
36254 36258 36260 36262 36265 36272 36283 36285 
36288 36297 36298 36306 36308 36325 36333 36342 
36353 36363 36366 36392 36397 36402 36407 36411 
36419 36422 36431 36440 36456 36459 36462 36465 
36469 36471 36473 36477 36478 36480 36483 36507 
36514 36519 36525 36526 36531 36532 36533 36534 
36535 36536 36537 36539 36540 36543 36545 36547 
36548 36551 36560 36565 36566 36569 36570 36580 
36586 36589 36592 36594 36595 36612 36623 36641 
36642 36648 36651 36653 36656 36657 36666 36668 
36669 36673 36682 36709 36714 36736 36748 36754 
36765 36768 36769 36770 36772 36773 36778 36780 
36787 36789 36792 36800 36803 36810 36812 36815 
36839 36844 36849 36871 36888 36901 36904 36905 
36906 36907 36908 36912 36915 36919 36922 36928 
36931 36934 36936 36940 36954 36959 36964 36966 
36970 36972 36977 37004 37006 37010 37014 37018 
37020 37028 37033 37035 37037 37047 37052 37056 
37058 37062 37065 37068 37069 37074 37102 37110 
37130 37132 37139 37157 37175 37180 37181 37186 
37201 37209 37211 37212 37222 37223 37227 37229 
37243 37244 37256 37262 37268 37269 37270 37272 
37286 37289 37302 37307 37311 37316 37330 37342 
37344 37349 37359 37360 37362 37370 37371 37374 
37384 37387 37395 37400 37403 37405 37407 37408 
37409 37410 37416 37418 37419 37423 37429 37435 
37436 37441 37442 37443 37447 37461 37464 37465 
37468 37469 37471 37480 37481 37482 37483 37486 
37491 37493 37495 37505 37508 37513 37519 37520 
37522 37534 37535 37551 37552 37553 37560 37562 
37565 37566 37567 37588 37590 37594 37595 37596 
37602 37603 37605 37611 37612 37620 37621 37622 
37629 37635 37637 37639 37642 37649 37655 37660 
37680 37681 37687 37693 37695 37696 37697 37698 
37699 37701 37715 37725 37727 37730 37734 37736 
37737 37739 37743 37746 37747 37748 37752 37757 
37759 37761 37764 37765 37766 37767 37771 37776 
37779 37788 37792 37803 37814 37816 37818 37819 
37820 37821 37822 37823 37825 37829 37830 37833 
37835 37843 37851 37856 37865 37866 37867 37869 
37871 37872 37873 37874 37875 37876 37889 37890 
37893 37896 37911 37915 37916 37917 37918 37919 
37922 37923 37924 37926 37927 37933 37935 37940 
37954 37955 37965 37966 37972 37974 37976 37977 
37980 37983 37985 37989 37990 37991 37996 38009 
38010 38011 38020 38033 38038 38040 38087 38095 
38099 38100 38106 38118 38139 38172 38176 38183 
38195 38205 38211 38216 38219 38229 38234 38240 
38254 38260 38261 38264 38265 38270 38273 38276 
38277 38293 38295 38298 38301 38310 38314 38319 
38323 38324 38328 38337 38338 38340 38350 38351 
38359 38361 38365 38374 38375 38380 38399 38407 
38419 38424 38427 38437 38438 38439 38441 38443 
38455 38456 38462 38465 38482 38486 38487 38489 
38490 38496 38510 38521 38523 38527 38529 38530 
38540 38550 38554 38557 38559 38565 38566 38571 
38573 38575 38581 38586 38608 38628 38630 38631 
38636 38637 38638 38644 38652 38659 38668 38676 
38677 38679 38682 38683 38689 38705 38707 38708 
38710 38711 38715 38716 38720 38721 38725 38730 
38732 38733 38734 38735 38736 38740 38741 38743 
38749 38751 38755 38759 38767 38773 38791 38793 
38796 38803 38805 38806 38811 38815 38823 38825 
38832 38833 38837 38840 38842 38844 38845 38846 
38848 38850 38858 38865 38866 38874 38875 38880 
38882 38884 38888 38895 38900 38903 38908 38921 
38923 38932 38933 38937 38943 38946 38947 38949 
38954 38958 38961 38963 38966 38970 38974 38975 
38976 38978 38983 38998 39009 39014 39016 39020 
39021 39022 39051 39054 39058 39061 39065 39075 
39083 39088 39092 39093 39095 39097 39109 39112 
39114 39117 39119 39120 39124 39126 39127 39133 
39136 39140 39142 39148 39152 39153 39155 39157 
39159 39160 39163 39167 39169 39172 39174 39179 
39182 39183 39193 39196 39202 39203 39206 39220 
39222 39223 39224 39225 39227 39232 39236 39238 
39242 39247 39258 39261 39264 39266 39268 39270 
39283 39288 39289 39291 39294 39298 39299 39305 
39308 39310 39322 39323 39327 39328 39330 39332 
39337 39338 39350 39351 39352 39358 39359 39360 
39370 39386 39390 39392 39393 39398 39400 39403 
39407 39411 39413 39424 39436 39440 39447 39448 
39455 39457 39462 39471 39475 39483 39484 39495 
39499 39505 39512 39516 39517 39521 39523 39538 
39555 39561 39565 39566 39572 39573 39577 39590 
39593 39594 39596 39598 39602 39605 39613 39619 
39624 39625 39630 39639 39641 39642 39643 39645 
39648 39652 39656 39657 39664 39669 39672 39679 
39680 39682 39687 39689 39700 39707 39708 39709 
39713 39718 39724 39725 39728 39732 39734 39736 
39737 39741 39744 39760 39763 39767 39772 39773 
39774 39778 39779 39781 39785 39786 39787 39789 
39790 39794 39795 39800 39801 39807 39809 39812 
39817 39818 39819 39820 39821 39823 39828 39832 
39833 39836 39837 39843 39847 39849 39852 39856 
39857 39858 39859 39863 39866 39867 39868 39870 
39874 39877 39883 39884 39885 39886 39888 39896 
39901 39903 39913 39917 39918 39919 39923 39924 
39926 39929 39930 39931 39932 39934 39935 39936 
39937 39938 39939 39951 39953 39958 39960 39961 
39962 39966 39967 39968 39970 39974 39975 39978 
39989 39992 40002 40003 40005 40015 40017 40019 
40027 40028 40029 40033 40036 40037 40041 40042 
40043 40044 40047 40048 40050 40059 40061 40062 
40064 40067 40068 40073 40074 40076 40079 40083 
40086 40087 40088 40089 40093 40106 40108 40111 
40121 40126 40127 40128 40129 40130 40136 40137 
40145 40146 40154 40155 40160 40161 40163 40164 
40168 40174 40175 40184 40190 40193 40194 40202 
40203 40204 40205 40207 40209 40211 40214 40218 
40220 40225 40228 40231 40242 40244 40245 40249 
40250 40252 40263 40265 40269 40270 40277 40290 
40291 40293 40294 40301 40302 40310 40316 40318 
40323 40333 40334 40337 40339 40341 40357 40366 
40368 40381 40384 40393 40404 40416 40423 40426 
40433 40444 40456 40458 40460 40462 40470 40472 
40476 40484 40487 40494 40496 40500 40507 40508 
40512 40525 40528 40530 40531 40532 40534 40537 
40541 40543 40544 40545 40546 40549 40558 40559 
40562 40564 40566 40567 40571 40580 40581 40591 
40592 40598 40600 40606 40610 40611 40616 40619 
40620 40623 40625 40626 40627 40641 40645 40646 
40647 40650 40651 40663 40673 40675 40684 40689 
40696 40706 40707 40708 40709 40716 40721 40724 
40733 40735 40742 40743 40762 40764 40767 40773 
40787 40794 40808 40809 40813 40819 40828 40829 
40833 40834 40846 40847 40851 40854 40855 40862 
40865 40867 40869 ])

(define-ccl-program ccl-encode-unicode-font
  `(0
    ;; In:  R0:charset id
    ;;      R1:position code 1
    ;;      R2:position code 2
    ;; Out: R1:font code point 1
    ;;      R2:font code point 2
    ((if (r0 == ,(charset-id 'latin-iso8859-1))
	 ((r2 = (r1 + 128))
	  (r1 = 0))
       (if (r0 == ,(charset-id 'lao))
	   ((r2 = (r1 + ?\x60))
	    (r1 = ?\x0e))
	 (if (r0 == ,(charset-id 'thai-tis620))
	     ((r2 = (r1 - ?\x20))
	      (r1 = ?\x0e))
	   ((r2 = ((((r1 - ?\x20) * 96) + r2) - ?\x20))
	    ;; r2 = index between 0x0000 and 0x2400
	    (if (r0 == ,(charset-id 'unicode-mono))
		;; Unicode-Mono: 0100-0dff,13a0-18ff,1e00-28ff,fb00-ffff
		((r2 += ?\x0100)
		 (if (r2 > ?\x0dff) (r2 += ?\x0200))
		 (if (r2 > ?\x10ff) (r2 += ?\x02a0))
		 (if (r2 > ?\x18ff) (r2 += ?\x0500))
		 (if (r2 > ?\x28ff) (r2 += ?\xd200)))
	      ;; Unicode-wide
	      (r2 = r2 ,mule-uni-1-table))
	    (r1 = (r2 / ?\x100))
	    (r2 %= ?\x100)))))))
  "CCL program to encode a char code to code point of Unicode font.")
  
(define-ccl-program ccl-encode-johab-font
  `(0
    ;; In:  R0:charset id (not checked)
    ;;      R1:position code 1
    ;;      R2:position code 2
    ;; Out: R1:font code point 1
    ;;      R2:font code point 2
    ((r2 = ((((r1 - ?\x20) * 96) + r2) - ?\x20))
     ;; r2 = index between 0x0000 and 0x2400
     (r1 = (r2 / ?\x100))
     (r2 %= ?\x100)))
  "CCL program to encode an char code to code point of Johab font.")

(setq font-ccl-encoder-alist
      (cons (cons "johab" ccl-encode-johab-font)
	    (cons (cons "iso10646" ccl-encode-unicode-font)
		  font-ccl-encoder-alist)))
  
(modify-coding-system-alist 'file "\\.utf\\'" 'unicode-utf8)
(modify-coding-system-alist 'file "\\.utf8\\'" 'unicode-utf8)
(modify-coding-system-alist 'file "\\.utf-8\\'" 'unicode-utf8)
(modify-coding-system-alist 'file "\\.u8\\'" 'unicode-utf8)

(setq x-pixel-size-width-font-regexp
      "gb2312\\|jisx0208\\|ksc5601\\|cns11643\\|big5\\|johab")



;; LIMITED Mac Roman support.  characters 1-159 are supported correctly.

;; Definitions for the Mac Roman character sets and coding system.
;; The Mac Roman encoding uses all 128 code points in the range 128 to
;; 255 for actual characters.  Emacs decodes them to one of the
;; following character sets.
;;	ascii, latin-iso8859-1, mule-unicode-0100-24ff,
;;	mule-unicode-2500-33ff, mule-unicode-e000-ffff

(defun decode-char (ccs code-point &optional restriction)
  "Return character specified by coded character set CCS and CODE-POINT in it.
Return nil if such a character is not supported.
Currently the only supported coded character set is `ucs' (ISO/IEC
10646: Universal Multi-Octet Coded Character Set).

Optional argument RESTRICTION specifies a way to map the pair of CCS
and CODE-POINT to a character.  Currently not supported and just ignored."
  (cond ((eq ccs 'ucs)
	 (cond ((< code-point 160)
		code-point)
	       ((< code-point 256)
		(make-char 'latin-iso8859-1 code-point))
 	       ((< code-point ?\x2500)
 		(setq code-point (- code-point ?\x0100))
 		(make-char 'unicode-mono  ; mule-unicode-0100-24ff
 			   (+ (/ code-point 96) 32) (+ (% code-point 96) 32)))
 	       ((< code-point ?\x3400)
 		(setq code-point (- code-point ?\x2500))
 		(make-char 'unicode-wide ; mule-unicode-2500-33ff
 			   (+ (/ code-point 96) 32) (+ (% code-point 96) 32)))
 	       ((and (>= code-point ?\xe000) (< code-point ?\x10000))
 		(setq code-point (- code-point ?\xe000))
 		(make-char 'unicode-mono ; mule-unicode-e000-ffff
 			   (+ (/ code-point 96) 32) (+ (% code-point 96) 32)))
	       (t (make-char 'unicode-mono code-point))
;	       (t (setq code-point (- code-point ?\x0100))
;		  (make-char 'unicode-mono (+ (/ code-point 96) 32) (+ (% code-point 96) 32)))
	       ))))


(let
    ((encoding-vector (make-vector 256 nil))
     (i 0)
     (vec	;; mac-roman (128..255) -> UCS mapping
      [ ?\x00C4	;; 128:LATIN CAPITAL LETTER A WITH DIAERESIS
	?\x00C5	;; 129:LATIN CAPITAL LETTER A WITH RING ABOVE
	?\x00C7	;; 130:LATIN CAPITAL LETTER C WITH CEDILLA
	?\x00C9	;; 131:LATIN CAPITAL LETTER E WITH ACUTE
	?\x00D1	;; 132:LATIN CAPITAL LETTER N WITH TILDE
	?\x00D6	;; 133:LATIN CAPITAL LETTER O WITH DIAERESIS
	?\x00DC	;; 134:LATIN CAPITAL LETTER U WITH DIAERESIS
	?\x00E1	;; 135:LATIN SMALL LETTER A WITH ACUTE
	?\x00E0	;; 136:LATIN SMALL LETTER A WITH GRAVE
	?\x00E2	;; 137:LATIN SMALL LETTER A WITH CIRCUMFLEX
	?\x00E4	;; 138:LATIN SMALL LETTER A WITH DIAERESIS
	?\x00E3	;; 139:LATIN SMALL LETTER A WITH TILDE
	?\x00E5	;; 140:LATIN SMALL LETTER A WITH RING ABOVE
	?\x00E7	;; 141:LATIN SMALL LETTER C WITH CEDILLA
	?\x00E9	;; 142:LATIN SMALL LETTER E WITH ACUTE
	?\x00E8	;; 143:LATIN SMALL LETTER E WITH GRAVE
	?\x00EA	;; 144:LATIN SMALL LETTER E WITH CIRCUMFLEX
	?\x00EB	;; 145:LATIN SMALL LETTER E WITH DIAERESIS
	?\x00ED	;; 146:LATIN SMALL LETTER I WITH ACUTE
	?\x00EC	;; 147:LATIN SMALL LETTER I WITH GRAVE
	?\x00EE	;; 148:LATIN SMALL LETTER I WITH CIRCUMFLEX
	?\x00EF	;; 149:LATIN SMALL LETTER I WITH DIAERESIS
	?\x00F1	;; 150:LATIN SMALL LETTER N WITH TILDE
	?\x00F3	;; 151:LATIN SMALL LETTER O WITH ACUTE
	?\x00F2	;; 152:LATIN SMALL LETTER O WITH GRAVE
	?\x00F4	;; 153:LATIN SMALL LETTER O WITH CIRCUMFLEX
	?\x00F6	;; 154:LATIN SMALL LETTER O WITH DIAERESIS
	?\x00F5	;; 155:LATIN SMALL LETTER O WITH TILDE
	?\x00FA	;; 156:LATIN SMALL LETTER U WITH ACUTE
	?\x00F9	;; 157:LATIN SMALL LETTER U WITH GRAVE
	?\x00FB	;; 158:LATIN SMALL LETTER U WITH CIRCUMFLEX
	?\x00FC	;; 159:LATIN SMALL LETTER U WITH DIAERESIS
	?\x2020	;; 160:DAGGER
	?\x00B0	;; 161:DEGREE SIGN
	?\x00A2	;; 162:CENT SIGN
	?\x00A3	;; 163:POUND SIGN
	?\x00A7	;; 164:SECTION SIGN
	?\x2022	;; 165:BULLET
	?\x00B6	;; 166:PILCROW SIGN
	?\x00DF	;; 167:LATIN SMALL LETTER SHARP S
	?\x00AE	;; 168:REGISTERED SIGN
	?\x00A9	;; 169:COPYRIGHT SIGN
	?\x2122	;; 170:TRADE MARK SIGN
	?\x00B4	;; 171:ACUTE ACCENT
	?\x00A8	;; 172:DIAERESIS
	?\x2260	;; 173:NOT EQUAL TO
	?\x00C6	;; 174:LATIN CAPITAL LETTER AE
	?\x00D8	;; 175:LATIN CAPITAL LETTER O WITH STROKE
	?\x221E	;; 176:INFINITY
	?\x00B1	;; 177:PLUS-MINUS SIGN
	?\x2264	;; 178:LESS-THAN OR EQUAL TO
	?\x2265	;; 179:GREATER-THAN OR EQUAL TO
	?\x00A5	;; 180:YEN SIGN
	?\x00B5	;; 181:MICRO SIGN
	?\x2202	;; 182:PARTIAL DIFFERENTIAL
	?\x2211	;; 183:N-ARY SUMMATION
	?\x220F	;; 184:N-ARY PRODUCT
	?\x03C0	;; 185:GREEK SMALL LETTER PI
	?\x222B	;; 186:INTEGRAL
	?\x00AA	;; 187:FEMININE ORDINAL INDICATOR
	?\x00BA	;; 188:MASCULINE ORDINAL INDICATOR
	?\x03A9	;; 189:GREEK CAPITAL LETTER OMEGA
	?\x00E6	;; 190:LATIN SMALL LETTER AE
	?\x00F8	;; 191:LATIN SMALL LETTER O WITH STROKE
	?\x00BF	;; 192:INVERTED QUESTION MARK
	?\x00A1	;; 193:INVERTED EXCLAMATION MARK
	?\x00AC	;; 194:NOT SIGN
	?\x221A	;; 195:SQUARE ROOT
	?\x0192	;; 196:LATIN SMALL LETTER F WITH HOOK
	?\x2248	;; 197:ALMOST EQUAL TO
	?\x2206	;; 198:INCREMENT
	?\x00AB	;; 199:LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
	?\x00BB	;; 200:RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
	?\x2026	;; 201:HORIZONTAL ELLIPSIS
	?\x00A0	;; 202:NO-BREAK SPACE
	?\x00C0	;; 203:LATIN CAPITAL LETTER A WITH GRAVE
	?\x00C3	;; 204:LATIN CAPITAL LETTER A WITH TILDE
	?\x00D5	;; 205:LATIN CAPITAL LETTER O WITH TILDE
	?\x0152	;; 206:LATIN CAPITAL LIGATURE OE
	?\x0153	;; 207:LATIN SMALL LIGATURE OE
	?\x2013	;; 208:EN DASH
	?\x2014	;; 209:EM DASH
	?\x201C	;; 210:LEFT DOUBLE QUOTATION MARK
	?\x201D	;; 211:RIGHT DOUBLE QUOTATION MARK
	?\x2018	;; 212:LEFT SINGLE QUOTATION MARK
	?\x2019	;; 213:RIGHT SINGLE QUOTATION MARK
	?\x00F7	;; 214:DIVISION SIGN
	?\x25CA	;; 215:LOZENGE
	?\x00FF	;; 216:LATIN SMALL LETTER Y WITH DIAERESIS
	?\x0178	;; 217:LATIN CAPITAL LETTER Y WITH DIAERESIS
	?\x2044	;; 218:FRACTION SLASH
	?\x20AC	;; 219:EURO SIGN
	?\x2039	;; 220:SINGLE LEFT-POINTING ANGLE QUOTATION MARK
	?\x203A	;; 221:SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
	?\xFB01	;; 222:LATIN SMALL LIGATURE FI
	?\xFB02	;; 223:LATIN SMALL LIGATURE FL
	?\x2021	;; 224:DOUBLE DAGGER
	?\x00B7	;; 225:MIDDLE DOT
	?\x201A	;; 226:SINGLE LOW-9 QUOTATION MARK
	?\x201E	;; 227:DOUBLE LOW-9 QUOTATION MARK
	?\x2030	;; 228:PER MILLE SIGN
	?\x00C2	;; 229:LATIN CAPITAL LETTER A WITH CIRCUMFLEX
	?\x00CA	;; 230:LATIN CAPITAL LETTER E WITH CIRCUMFLEX
	?\x00C1	;; 231:LATIN CAPITAL LETTER A WITH ACUTE
	?\x00CB	;; 232:LATIN CAPITAL LETTER E WITH DIAERESIS
	?\x00C8	;; 233:LATIN CAPITAL LETTER E WITH GRAVE
	?\x00CD	;; 234:LATIN CAPITAL LETTER I WITH ACUTE
	?\x00CE	;; 235:LATIN CAPITAL LETTER I WITH CIRCUMFLEX
	?\x00CF	;; 236:LATIN CAPITAL LETTER I WITH DIAERESIS
	?\x00CC	;; 237:LATIN CAPITAL LETTER I WITH GRAVE
	?\x00D3	;; 238:LATIN CAPITAL LETTER O WITH ACUTE
	?\x00D4	;; 239:LATIN CAPITAL LETTER O WITH CIRCUMFLEX
	?\xF8FF	;; 240:Apple logo
	?\x00D2	;; 241:LATIN CAPITAL LETTER O WITH GRAVE
	?\x00DA	;; 242:LATIN CAPITAL LETTER U WITH ACUTE
	?\x00DB	;; 243:LATIN CAPITAL LETTER U WITH CIRCUMFLEX
	?\x00D9	;; 244:LATIN CAPITAL LETTER U WITH GRAVE
	?\x0131	;; 245:LATIN SMALL LETTER DOTLESS I
	?\x02C6	;; 246:MODIFIER LETTER CIRCUMFLEX ACCENT
	?\x02DC	;; 247:SMALL TILDE
	?\x00AF	;; 248:MACRON
	?\x02D8	;; 249:BREVE
	?\x02D9	;; 250:DOT ABOVE
	?\x02DA	;; 251:RING ABOVE
	?\x00B8	;; 252:CEDILLA
	?\x02DD	;; 253:DOUBLE ACUTE ACCENT
	?\x02DB	;; 254:OGONEK
	?\x02C7	;; 255:CARON
	])
     translation-table)
  (while (< i 128)
    (aset encoding-vector i i)
    (setq i (1+ i)))
  (while (< i 256)
    (aset encoding-vector i
	  (decode-char 'ucs (aref vec (- i 128))))
    (setq i (1+ i)))
  (setq translation-table
	(make-translation-table-from-vector encoding-vector))
  (define-translation-table 'mac-roman-decoder translation-table)
  (define-translation-table 'mac-roman-encoder
    (char-table-extra-slot translation-table 0)))

(define-ccl-program decode-mac-roman
  `(4
    ((loop
      (read r1)
      (if (r1 < 128)  ;; ASCII
	  (r0 = ,(charset-id 'ascii))
	(if (r1 < 160)
	    (r0 = ,(charset-id 'ascii)) ;; eight-bit-control
	  (r0 = , (charset-id 'ascii)))) ;; eight-bit-graphic
      (translate-character mac-roman-decoder r0 r1)
      (write-multibyte-character r0 r1)
      (repeat))))
  "CCL program to decode Mac Roman")

(define-ccl-program encode-mac-roman
  `(1
    ((loop
      (read-multibyte-character r0 r1)
      (translate-character mac-roman-encoder r0 r1)
      (if (r0 != ,(charset-id 'ascii))
	  (if (r0 != ,(charset-id 'unicode-mono)) ;; eight-bit-graphic
	      (if (r0 != ,(charset-id 'unicode-wide)) ;; eight-bit-control
		  (r1 = ??))))
      (write-repeat r1))))
  "CCL program to encode Mac Roman")

(make-coding-system
 'mac-roman 4 ?M
 "Mac Roman Encoding"
 '(decode-mac-roman . encode-mac-roman)
 (list '(safe-charsets . t)
; (cons 'safe-chars (get 'mac-roman-encoder 'translation-table))
       '(valid-codes (0 . 255))
       '(mime-charset . macintosh)))
(define-coding-system-alias 'mac-roman-unix 'mac-roman)

(provide 'utf8)

;;; utf8.el ends here
