;;; ns-startup.el --- startup lisp for NeXT/Open/GNUstep/MacOS X window system
;; Copyright (C) 1993 Free Software Foundation, Inc.

;; Author: FSF
;; Keywords: terminals

;;; This file is part of GNU Emacs.
;;;
;;; GNU Emacs is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; GNU Emacs is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to
;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;; Convenience definitions
;; These are not strictly necessary for ns-emacs startup but are included
;; for user convenience.  They can be overridden in .emacs .

(require 'utf8)

;; Most of these are in fact set up by default in emacs-21.
(if (eq emacs-major-version '20) ((lambda ()

    ;;; Mouse wheel: wheel events are reported by emacs as
    ;;;              button 4 (scroll down) or button 5 (scroll up).
    ;;; TODO: document these and make available for easy customization
    (defun up-slightly () (interactive) (scroll-up 5))
    (defun down-slightly () (interactive) (scroll-down 5))
    (global-set-key [mouse-4] 'down-slightly)
    (global-set-key [mouse-5] 'up-slightly)

    (defun up-one () (interactive) (scroll-up 1))
    (defun down-one () (interactive) (scroll-down 1))
    (global-set-key [S-mouse-4] 'down-one)
    (global-set-key [S-mouse-5] 'up-one)

    ;; alt-up/down scrolling a la Stuart.app
    (global-set-key [M-up] 'down-one)
    (global-set-key [M-down] 'up-one)

    ;; Better Objective-C mode than in the stock cc-mode.
    (require 'objc-c-mode_20)

    ;; tramp remote file access (bundled with emacs-21)
    (require 'tramp)
    (setq tramp-default-method "scp")
    (setq tramp-auto-save-directory "/tmp")

    ;; turn on iswitch and imenu?  might surprise some, but it's good for you
    (iswitchb-default-keybindings)
    (add-hook 'c-mode-common-hook '(lambda () (imenu-add-menubar-index)))

    ;; Resizable minibuffer (who wouldn't want this?)
    (require 'rsz-mini)
    (resize-minibuffer-mode 1)

    ;; For some reason this gets overwritten by ns-menu-bar.el
    (load-library "mule-cmds")

)) )

;; Set up browser connectivity for OS X and GNUstep (move to ns-win.el?)
(setq browse-url-browser-function 'browse-url-generic)
(cond ((eq system-type 'apple-macosx)
       (setq browse-url-generic-program "open"))
      ;; otherwise, gnustep
      (t
       (setq browse-url-generic-program "gopen")) )

; Set up bundled ispell on OS X
(if (eq system-type 'apple-macosx)
    (setq ispell-extra-args (list "-d" (concat exec-directory "lib/americanmed"))) )


; page-down,page-up etc. leaves cursor in same place

;; ----------------------------------------------------------------------
;;; from pager.el --- windows-scroll commands
;;; Version 2.0 - 97-10-06
;;; Copyright (C) 1992-1997 Mikael Sjodin (mic@docs.uu.se)
;;; http://user.it.uu.se/~mic/emacs.shtml

(defvar pager-temporary-goal-column 0
  "Similar to temporary-goal-column but used by the pager.el functions")
;(make-variable-buffer-local 'pager-temporary-goal-column)

(defconst pager-keep-column-commands
  '(pager-row-down pager-row-up row-dn row-up
		   pager-page-down pager-page-up pg-dn pg-up)
  "Commands which when called without any other intervening command should
keep the `pager-temporary-goal-column'")

(defun pager-page-down ()
  "Like scroll-up, but moves a fixed amount of lines (fixed relative the
`window-height') so that pager-page-up moves back to the same line."
  (interactive)
  (if (not (pos-visible-in-window-p (point-max)))
      (pager-scroll-screen (- (1- (window-height))
			      next-screen-context-lines))))

(defun pager-page-up ()
  "Like scroll-down, but moves a fixed amount of lines (fixed relative the
`window-height') so that pager-page-down moves back to the same line."
  (interactive)
  (if (not (pos-visible-in-window-p (point-min)))
      (pager-scroll-screen (- next-screen-context-lines
			      (1- (window-height))))))

;; ------------------------------

(defun pager-scroll-screen (lines)
  "Scroll screen LINES, but keep the cursors position on screen."
  (if (not (memq last-command pager-keep-column-commands))
      (setq pager-temporary-goal-column (current-column)))
  (save-excursion
    (goto-char (window-start))
    (forward-line lines)
    (set-window-start (selected-window) (point)))
  (forward-line lines)
  (move-to-column pager-temporary-goal-column))

(global-set-key "\C-v"     'pager-page-down)
(global-set-key [next]     'pager-page-down)
(global-set-key "\ev"      'pager-page-up)
(global-set-key [prior]    'pager-page-up)

;; note: the following only preserves line, not column
;(setq scroll-preserve-screen-position t)


;; ----------------------------------------------------------------------
;; mic-paren: improved paren matching allowing sexp to be highlighted in
;; more subdued color than the parens themselves
(load "mic-paren")
(paren-activate)
(setq paren-sexp-mode t)
(setq paren-match-face 'paren-face-match-light)
(defface paren-face-match-light
  '((((class color)) (:background "ARGBffd8eeee"))
    (t (:background "gray")))
  "Mic-paren mode face used for a matching paren."
  :group 'faces
  :group 'mic-paren-matching)

;; Load enhanced mark navigation commands
(load "mark-nav")

;; Load grabenv
(autoload 'grabenv "grabenv" "Get environment from your shell." t nil)

;; Whitespace.el from 2005/09/15 unicode-2 emacs CVS
(autoload 'whitespace-toggle-global-mode "whitespace" "Whitespace." t nil)
(autoload 'whitespace-toggle-leading-check "whitespace" "Whitespace." t nil)
(autoload 'whitespace-toggle-trailing-check "whitespace" "Whitespace." t nil)
(autoload 'whitespace-toggle-indent-check "whitespace" "Whitespace." t nil)
(autoload 'whitespace-toggle-spacetab-check "whitespace" "Whitespace." t nil)
(autoload 'whitespace-toggle-ateol-check "whitespace" "Whitespace." t nil)
(autoload 'whitespace-buffer "whitespace" "Whitespace." t nil)
(autoload 'whitespace-region "whitespace" "Whitespace." t nil)
(autoload 'whitespace-cleanup "whitespace" "Whitespace." t nil)
(autoload 'whitespace-cleanup-region "whitespace" "Whitespace." t nil)
(defalias 'global-whitespace-mode 'whitespace-global-mode)
(defvar whitespace-global-mode nil "Whitespace.")
(put 'whitespace-global-mode 'custom-set 'custom-set-minor-mode)
(autoload 'whitespace-global-mode "whitespace" "Whitespace." t nil)
(autoload 'whitespace-write-file-hook "whitespace" "Whitespace." t nil)

;; Temporarily work around bug in NS Emacs..
(setq ediff-window-setup-function 'ediff-setup-windows-plain)
