/* Character conversion from Emacs "Mule" internal coding to UTF-8 / Unicode.
   Copyright (C) 1989, 1993, 2005 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Original author: Otfreid Cheong (Copyright (c) 1999, 2000)
Adapted by Adrian Robert (arobert@cogsci.ucsd.edu), 2005

*/

#include "config.h"
#include "lisp.h"
#include "charset.h"
#include "nscharset_data.h"
#include <stdio.h>

//#define NSCHARSET_DEBUG

/* Internal helper routines. */
void makeUni2MuleMap();

static inline int ucsToUtf8(int charset, int ind, char *outpos)
{
    UCS4 ch;
    char *start = outpos;

    if (mule2uni[charset] && mule2uni[charset][ind]) {
      ch = mule2uni[charset][ind];
      if (ch < 0x80) {
        *outpos++ = ch;
      } else if (ch < 0x800) {
        *outpos++     = (((ch & 0x7c0) >> 6) | 0xc0);
        *outpos++ = (ch & 0x03f) | 0x80;
      } else {
          *outpos++ = ((ch & 0x0f000) >> 12) | 0xe0;
          *outpos++ = ((ch & 0xfc0) >> 6) | 0x80;
          *outpos++ = (ch & 0x03f) | 0x80;
      }
    } else if (charset == 0x7d) {
#ifdef THAI_LAO_KOREAN
        JohabToJamo(ind);
#else
#ifdef NSCHARSET_DEBUG
      fprintf(stderr, "UCS: korean johab/jamo charset %d\n", charset);
#endif
      *outpos++ = '.';  
#endif
    } else {
#ifdef NSCHARSET_DEBUG
      fprintf(stderr, "UCS: unrecog charset %d\tchar %d\n",
              charset, ind);
#endif
      *outpos++ = '.';
    }
    return outpos - start;
}


/* This function used by nswrap_moveshow for character rendering.
   It needs to be fast. */
void muleToUTF8s(const int mule_chs[], int size, unsigned char *outbuf)
{
    int charset, b1, b2, ind;
    int i;
    unsigned char *outpos = outbuf;
    char c2 = 0;

    for (i=0; i<size; i++) {

        if (mule_chs[i] < 0x80) {
            *outpos++ = mule_chs[i];
            continue;
        }

        SPLIT_NON_ASCII_CHAR(mule_chs[i], charset, b1, b2);/* from charset.h */
        charset -= 0x80;

        if (charset == 0) {
            /* composition  not handled yet; see xterm.c line 590 */
            *outpos++ = '.';
#ifdef NSCHARSET_DEBUG
            fprintf(stderr, "Mule: composition not handled yet for %d\n",
                    mule_chs[i]);
#endif
            continue;
        }

        b1 = (b1 | 0x80) - 0xa0;
        b2 = (b2 | 0x80) - 0xa0;
        switch (Mule_dimension[charset]) {
        case 1:
            ind = b1;
            break;
        case 2:
            ind = b1 * 0x60 + b2;
            i++; // emacs allocates two glyphs for 2-byte chars... (?)
            c2 = 1;
            break;
        default:
#ifdef NSCHARSET_DEBUG
            fprintf(stderr, "Mule: no translation for charset %d\n", charset);
#endif
            *outpos++ = '.';
            continue;
        }
/* fprintf(stderr, "charset=%d, ind=%d\t(%d, %d)\n",charset,ind,(ind>>8)&0xff, (ind&0xff)); */

        /* at this point we have a UCS character; now go to UTF-8 */
        outpos += ucsToUtf8(charset, ind, outpos);

        // (temporary?) hack
        if (c2) {
            *outpos++=' ';
            c2 = 0;
        }
    } /* for over mule_chs */

    *outpos++ = '\0';
}


/* Used for pasteboard out. */
void mulemToUTF8s(const char *mule_str, int size, char *outbuf)
{
    int charset, b1, b2, ind;
    const unsigned char *inpos = mule_str;
    char *outpos = outbuf;
    int with_rule = 0;
    UCS4 p;

    while (inpos - (unsigned char *)mule_str < size) {
        p = *inpos++;

        if (with_rule) {
          if (p < 0xa0) {
            /* end of composed character */
            with_rule = 0;
          } else {
            /* this is actually a composition rule, ignore it */
            p = *inpos++;
          }
        }

        if (p < 0x80 || p == EOF) {
          *outpos++ = p;
          continue;
        }

        /* skip composition prefix */
        if (p == 0x80) {
          p = *inpos++;
          if (p == 0xff) {
            /* Composing with rule */
            with_rule = 1;
            p = *inpos++;
          }
        }

        /* handle case of composed chars */
        if (p == 0xa0) {
          /* a composing Ascii character */
          *outpos++ = (*inpos++ - 0x80);
          continue;
        }
        if (p > 0xa0) {
          /* a composing character => decompose */
          p -= 0x20;
        }
        if (0x9a <= p && p <= 0x9d) {
          /* only prefix code, get real charset */
          p = *inpos++;
        }

        /* now p contains charset */
        charset = p - 0x80;
        b1 = *inpos++ - 0xa0;
        switch (Mule_dimension[charset]) {
        case 1:
          ind = b1;
          break;
        case 2:
          b2 = (*inpos++ & 0xff) - 0xa0;
          ind = b1 * 0x60 + b2;
          break;
        default:
#ifdef NSCHARSET_DEBUG
          fprintf(stderr, "Mule: no translation for charset %d\n", charset);
#endif
          *outpos++ = '.';
          continue;
        }

        outpos += ucsToUtf8(charset, ind, outpos);
    }
    *outpos++ = '\0';
}


/* Used for pasteboard in. */
void UTF8sToMulem(const char *utf_str, int size, char *outbuf)
{
    const unsigned char *inpos = utf_str;
    char *outpos = outbuf;
    UCS4 ch, ch2, mch;
    int extraBytes, charset, ind;
    int composition_offset = 0; // should be 0x20 for Thai/Lao composition...
    char cOuter;
    static char firstTime = 1;

    if (firstTime) {
        makeUni2MuleMap(); /* make a 4Mb array ;-( */
        firstTime = 0;
    }

    while (inpos - (unsigned char *)utf_str < size) {
        ch = *inpos++;

        if (ch == EOF || ch == 0xfe || ch == 0xff || (ch & 0xc0) == 0x80) {
#ifdef NSCHARSET_DEBUG
            fprintf(stderr, "UTF8: unexpected character %ld\n", ch);
#endif
            *outpos++ = '.';
            continue;
        }

        extraBytes = bytesFromUTF8[ch & 0xff];
        ch -= firstByteMark[extraBytes];
        cOuter = 0;
        while (extraBytes--) {
          ch <<= 6;
          ch2 = *inpos++;
          if (ch2 == EOF || (ch2 & 0xc0) != 0x80) {
#ifdef NSCHARSET_DEBUG
            fprintf(stderr, "UTF8: unexpected second character %ld\n", ch2);
#endif
            *outpos++ = '.';
            cOuter = 1;
            break;
          }
          if (cOuter)
              continue;
          ch |= ch2 & 0x3f;
        }

        if (ch >= 0x10000) {
#ifdef NSCHARSET_DEBUG
            fprintf(stderr, "UTF8: unexpected character %ld\n", ch);
#endif
            *outpos++ = '.';
            continue;
        }

        /* now ch is the UCS representation */
        if (ch < 0x80) {
            *outpos++ = ch;
            continue;
        }

        mch = uni2mule[ch];

#ifdef THAI_LAO_KOREAN
        if (0x0e00 <= ch && ch < 0x0f00 && mch) {
            /* Thai and Lao */
            FlushJamo();
            ComposeThaiLao(ch);
            continue;
        }
        FlushThaiLao();

        if (!mch) {
            if (0xac00 <= ch && ch <= 0xd7a3) {
                /* Korean Hangul Syllable */
                DecomposeHangul(ch);
            } else if (0x1100 <= ch && ch < 0x1200) {
                /* Korean Jamo */
                ComposeJamo(ch);
            } else {
                *outpos++ = '.';
#ifdef NSCHARSET_DEBUG
                fprintf(stderr, "UCS: unrecognized character: %ld\n", ch);
#endif
            }
            continue;
        }
        FlushJamo();
#else
        if (!mch) {
            *outpos++ = '.';
#ifdef NSCHARSET_DEBUG
            fprintf(stderr, "UCS: unrecognized character: %ld\n", ch);
#endif
            continue;
        }
#endif

        charset = ((mch & 0xff0000) >> 16) - 0x80;
        if (Mule_prefix[charset]) {
          *outpos++ = Mule_prefix[charset] + composition_offset;
          *outpos++ = charset + 0x80;
        } else {
          *outpos++ = charset + 0x80 + composition_offset;
        }
        switch (Mule_dimension[charset]) {
        case 1:
          *outpos++ = (mch & 0xffff) + 0xa0;
          break;
        case 2:
          ind = mch & 0xffff;
          *outpos++ = (ind / 0x60) + 0xa0;
          *outpos++ = (ind % 0x60) + 0xa0;
          break;
        default:
          break;
        }
    }
    *outpos++ = '\0';
}


/* utility */
void makeUni2MuleMap()
{
  int charset, i, j, uni;
  for (i = 0; i < 0x10000; i++)
    uni2mule[i] = 0x0;
  for (j = 0; j < sizeof(charset_priorities)/sizeof(int); j++) {
    charset = charset_priorities[j] - 0x80;
    if (mule2uni[charset]) {
      int size = (Mule_dimension[charset] == 2) ? 0x60*0x60 : 0x60;
      for (i = 0; i < size; i++) {
	uni = mule2uni[charset][i];
	if (uni && !uni2mule[uni]) 
	  uni2mule[uni] = (charset << 16) + 0x800000 + i;
      }
    }
  }
}
