/* NeXTstep menu module
   Copyright (C) 1986, 1988, 1993 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  

Originally by Carl Edman
Updated by Christian Limpach (chris@nice.ch)
OpenStep/Rhapsody  port by Scott Bender (sbender@harmony-ds.com)
MacOSX/Aqua port by Christophe de Dinechin (descubes@earthlink.net)
GNUstep port by Adrian Robert (arobert@cogsci.ucsd.edu)

*/

#import <AppKit/AppKit.h>

#include <signal.h>
#include "config.h"
#include "lisp.h"
#include "frame.h"
#include "buffer.h"
#include "window.h"
#include "keyboard.h"
#include "commands.h"
#include "blockinput.h"
#include "nsgui.h"
#include "nsterm.h"
#include "dispextern.h"
#include "termhooks.h"

//#include <sys/timeb.h>
//#include <sys/types.h>

#define MenuStagger 10.0

#if 0
int menu_trace_num = 0;
#define TRACE(x)        fprintf(stderr, "%s:%d: [%d] " #x "\n",        \
                                __FILE__, __LINE__, ++menu_trace_num)
#else
#define TRACE(x)
#endif

extern Lisp_Object Qundefined;
extern Lisp_Object Qmenu_enable;

Lisp_Object Vmenu_updating_frame;

Lisp_Object ns_menu_path;
EmacsMenu *mainMenu, *svcsMenu;


DEFUN ("x-popup-menu",Fx_popup_menu, Sx_popup_menu, 1, 2, 0,
       "Pop up a deck-of-cards menu and return user's selection.\n\
POSITION is a position specification.  This is either a mouse button event\n\
or a list ((XOFFSET YOFFSET) WINDOW)\n\
where XOFFSET and YOFFSET are positions in pixels from the top left\n\
corner of WINDOW's frame.  (WINDOW may be a frame object instead of a window.)\n\
This controls the position of the center of the first line\n\
in the first pane of the menu, not the top left of the menu as a whole.\n\
\n\
MENU is a specifier for a menu.  For the simplest case, MENU is a keymap.\n\
The menu items come from key bindings that have a menu string as well as\n\
a definition; actually, the \"definition\" in such a key binding looks like\n\
\(STRING . REAL-DEFINITION).  To give the menu a title, put a string into\n\
the keymap as a top-level element.\n\n\
You can also use a list of keymaps as MENU.\n\
  Then each keymap makes a separate pane.\n\
When MENU is a keymap or a list of keymaps, the return value\n\
is a list of events.\n\n\
Alternatively, you can specify a menu of multiple panes\n\
  with a list of the form (TITLE PANE1 PANE2...),\n\
where each pane is a list of form (TITLE ITEM1 ITEM2...).\n\
Each ITEM is normally a cons cell (STRING . VALUE);\n\
but a string can appear as an item--that makes a nonselectable line\n\
in the menu.\n\
With this form of menu, the return value is VALUE from the chosen item.")
     (position, menu)
     Lisp_Object position, menu;
{
  // NOTE: this also gets called when a choice is made in the app main menu;
  //       the second part of ns-menu-path is sent back
  id panel;
  Lisp_Object window,x,y,tem;
  struct frame *f;
  NSPoint p;

  TRACE(x-popup-menu);

  if (NILP(position))
    {
      /* Don't create a menu.  Just precalculate the keyboard equivalents */
      return Qnil;
    }
  
  if (CONSP(ns_menu_path))
    {
      Lisp_Object ret=XCONS(ns_menu_path)->car;
      ns_menu_path=XCONS(ns_menu_path)->cdr;
      return ret;
    }
  
  check_ns();
  
  if (EQ (position, Qt))
    {
      /* Use the mouse's current position.  */
      struct frame *new_f = 0;

      if (mouse_position_hook)
        (*mouse_position_hook) (&new_f, 0, 0, 0, &x, &y, 0);
      if (new_f != 0)
        XSETFRAME (window, new_f);
      else
        {
          window = selected_window;
          x = make_number(0);
          y = make_number(0);
        }
    }
  else
    {
      CHECK_CONS (position,0);
      tem = Fcar (position);
      if (XTYPE (tem) == Lisp_Cons)
        {
          window = Fcar (Fcdr (position));
          x = Fcar(tem);
          y = Fcar(Fcdr (tem));
        }
      else
        {
          tem = Fcar (Fcdr (position));
          window = Fcar (tem);
          tem = Fcar (Fcdr (Fcdr (tem)));
          x = Fcar (tem);
          y = Fcdr (tem);
        }
    }
  
  CHECK_NUMBER (x,0);
  CHECK_NUMBER (y,0);
  
  if (FRAMEP (window))
    {
      f = XFRAME (window);
      
      p.x = 0;
      p.y = 0;
    }
  else
    {
      CHECK_LIVE_WINDOW (window, 0);
      f = XFRAME (WINDOW_FRAME (XWINDOW (window)));
      
      p.x = ((int)(NS_FONT_WIDTH(f)) * XWINDOW (window)->left);
      p.y = (NS_FRAME_LINE_HEIGHT(f) * XWINDOW (window)->top);
    }
  
  p.x += XINT(x); p.y += XINT(y);

  p = [FRAME_NS_VIEW(f) convertPoint: p toView: nil];
  p = [[FRAME_NS_VIEW(f) window] convertBaseToScreen: p];
  panel=[[EmacsMenuPanel alloc] initFromMenu: menu];
  
  tem=[panel runMenuAt: p.x : p.y];

  //HACK to work around fact that on linux, 'close' is #defined to sys_close
  // in s/gnu-linux.h; problem will go away on emacs-21+
#if defined(__linux__) && defined(close)
#undef close
  [panel close];
#define close sys_close
#else
  [panel close];
#endif  // linux && close

  [[FRAME_NS_VIEW(selected_frame) window] makeKeyWindow];
  return tem;
}

DEFUN ("x-popup-dialog", Fx_popup_dialog, Sx_popup_dialog, 2, 2, 0,
       "Pop up a dialog box and return user's selection.\n\
POSITION specifies which frame to use.\n\
This is normally a mouse button event or a window or frame.\n\
If POSITION is t, it means to use the frame the mouse is on.\n\
The dialog box appears in the middle of the specified frame.\n\
\n\
CONTENTS specifies the alternatives to display in the dialog box.\n\
It is a list of the form (TITLE ITEM1 ITEM2...).\n\
Each ITEM is a cons cell (STRING . VALUE).\n\
The return value is VALUE from the chosen item.\n\n\
An ITEM may also be just a string--that makes a nonselectable item.\n\
An ITEM may also be nil--that means to put all preceding items\n\
on the left of the dialog box and all following items on the right.\n\
\(By default, approximately half appear on each side.)")
     (position, contents)
     Lisp_Object position, contents;
{
  id dialog;
  Lisp_Object window,tem;
  struct frame *f;
  NSPoint p;

  TRACE(x-popup-dialog);
  
  check_ns();
  if (EQ (position, Qt))
    {
      window = selected_window;
    }
  else if (CONSP (position))
    {
      Lisp_Object tem;
      tem = Fcar (position);
      if (XTYPE (tem) == Lisp_Cons)
        window = Fcar (Fcdr (position));
      else
        {
          tem = Fcar (Fcdr (position));  /* EVENT_START (position) */
          window = Fcar (tem);	     /* POSN_WINDOW (tem) */
        }
    }
  else if (FRAMEP (position))
    {
      window = position;
    }
  else
    {
      CHECK_LIVE_WINDOW (position,0);
      window = position;
    }
  
  if (FRAMEP (window))
    f = XFRAME (window);
  else
    {
      CHECK_LIVE_WINDOW (window, 0);
      f = XFRAME (WINDOW_FRAME (XWINDOW (window)));
    }
  p.x = (int)f->output_data.ns->left +
    ((int)(f->output_data.ns->font->width) * f->width)/2;
  p.y = (int)f->output_data.ns->top +
    (f->output_data.ns->line_height * f->height)/2;
  dialog=[[EmacsDialogPanel alloc] initFromContents: contents];
  
  tem=[dialog runDialogAt: p.x : p.y];

  //HACK to work around fact that on linux, 'close' is #defined to sys_close
  // in s/gnu-linux.h; problem will go away on emacs-21+
#if defined(__linux__) && defined(close)
#undef close
  [dialog close];
#define close sys_close
#else
  [dialog close];
#endif  // linux && close

  [[FRAME_NS_VIEW(selected_frame) window] makeKeyWindow];
  return tem;
}

void free_frame_menubar (struct frame *f)
{
  EmacsMenu *menu = (EmacsMenu *) [NSApp mainMenu];

  TRACE(free_frame_menubar);

  [menu mark];
  [menu freeMarked];
}

//  Lisp_Object Vns_menubar;

/* Called to indicate that the frame's menus have changed, and the *step
   representation should be updated from Lisp.
   Performance note- this function gets called with every scroll.
   Should find way to detect and return from frivolous calls, but for now
   just need to make sure processing is efficient.  The EmacsMenu marking
   mechanism causes this to be a sub-10msec call after the first init for
   a window.
*/
void set_frame_menubar (struct frame *f, int first_time, int deep_p)
{
  int i;
  Lisp_Object items,tail,key,string;
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
  BOOL needsSet = NO;
  id menu = [NSApp mainMenu];
  id attMenu;
//      struct timeb tb;
//      long t;

  TRACE(set_frame_menubar);
  //NSLog(@"got main menu: '%@'", menu);

  /** Menu may have been created automatically. **/
  if ([menu isKindOfClass:[EmacsMenu class]] == NO) 
    menu = nil;

  if (f != selected_frame)
    {
      [pool release];
      return;
    }

  if (menu == nil)
    {
      menu = [[EmacsMenu alloc] initWithTitle: @"Emacs"];
      // set menu to not be autoupdated by NSApp.. must do ourselves
      [menu setAutoenablesItems: NO];
      needsSet = YES;
    }

  // close up anything on there
  attMenu = [menu attachedMenu];
  if (attMenu != nil)
#if defined(__linux__) && defined(close)
#undef close
    [attMenu close];
#define close sys_close
#else
  [attMenu close];
#endif  // linux && close

  // make sure we're playing with the latest deck
  //safe_run_hooks (Qmenu_bar_update_hook); // check if necessary
  if (NILP (items=FRAME_MENU_BAR_ITEMS (f)))
    {
      // no point, will soon be called again, so just return
      //items= FRAME_MENU_BAR_ITEMS(f)=menu_bar_items(FRAME_MENU_BAR_ITEMS(f));
      [pool release];
      return;
    }
  items= FRAME_MENU_BAR_ITEMS(f)= menu_bar_items(FRAME_MENU_BAR_ITEMS(f));
//debug_print(items);
//    Vns_menubar = items;
//ftime(&tb);
//t = -(1000*tb.time+tb.millitm);

  [menu mark];

  // this initializes/updates the Emacs menus from LISP
  for (i=0; i<XVECTOR(items)->size; i+=4)
    {
      key=XVECTOR(items)->contents[i];
      string=XVECTOR(items)->contents[i+1];
      tail=XVECTOR(items)->contents[i+2];
      if (NILP (string))
        break;

    //PENDING: below is correct, but apparently something isn't being
    // checked for in adjustKey that leads to a segfault down the road
    //  for (tail=XVECTOR(items)->contents[i+2]; CONSP(tail); tail=XCONS(tail)->cdr){
#ifdef MACOSX
      if (!strcmp("Services",XSTRING(string)->data))
        {
          NSMenu *appMenu = [[mainMenu itemAtIndex: 0] submenu];
          NSMenuItem *cell =
            [appMenu itemAtIndex: [appMenu indexOfItemWithSubmenu:svcsMenu]];
          EmacsMenuObject *obj;
          if ([cell representedObject] == nil)
            {
              obj = [[[EmacsMenuObject alloc] init] autorelease];
              [obj setValue: key];
              [obj setSupercell: nil];
              [cell setRepresentedObject:obj];
            }
          [svcsMenu adjustMap: XCONS(tail)->car];
        }
      else
#endif
        [menu adjustKey:key binding:XCONS(tail)->car string:string];

      //PENDING: test whether we missed something (see PENDING above)
      for (tail=XCONS(tail)->cdr; CONSP(tail); tail=XCONS(tail)->cdr)
        fprintf(stderr, "ignoring top-level adjKey (please report as bug): '%s'\n",XSTRING(string)->data);

    }

//fprintf(stderr, "\nPRE free:\n");
//[menu dump: @"  "];

  [menu freeMarked];

//ftime(&tb);
//t += 1000*tb.time+tb.millitm;
//fprintf(stderr, "Menu update took %ld msec.\n",t);
//fprintf(stderr, "\nPOST free:\n");
//[menu dump: @"  "];

  // end init

  // set main menu
  if (needsSet)
    [NSApp setMainMenu: menu];

  [pool release];
}


DEFUN ("x-reset-menu",Fx_reset_menu, Sx_reset_menu, 0, 0, 0,
       "Cause the NS menu to be re-calculated.")
     ()
{
  set_frame_menubar(selected_frame,1, 0);
  return Qnil;
}

Lisp_Object ns_map_event_to_object (struct input_event *event, struct frame *f)
{
  TRACE(ns_map_event_to_object);
  if (CONSP(ns_menu_path))
    {
      Lisp_Object ret=XCONS(ns_menu_path)->car;
      ns_menu_path=XCONS(ns_menu_path)->cdr;
      return ret;
    }
  return Qnil;
}

void syms_of_nsmenu ()
{
  ns_menu_path=Qnil;
  staticpro(&ns_menu_path);
  defsubr (&Sx_popup_menu);
  defsubr (&Sx_popup_dialog);
  defsubr (&Sx_reset_menu);
  DEFVAR_LISP ("menu-updating-frame", &Vmenu_updating_frame,
    "Frame for which we are updating a menu.\n\
The enable predicate for a menu command should check this variable.");
  Vmenu_updating_frame = Qnil;
//    DEFVAR_LISP ("ns-menubar", &Vns_menubar,
//                 "Temp debug menubar storage.");
}

// link btwn NS menuitem and keymap lisp command object
@implementation EmacsMenuObject

- init
{
  [super init];
  value = Qnil;
  supercell = nil;
  return self;
}

- (void)dealloc
{
  [super dealloc]; 
  return;
}

- (Lisp_Object) value
{
  return value;
}

- setValue:(Lisp_Object)nvalue
{
  // note, care must be taken not to retain/use this ref past when lisp
  // garbage-collects it (and it cannot easily be protected from such)
  if (nvalue != value)
    value=nvalue;
  return self;
}

- supercell
{
  return supercell;
}

- setSupercell:(id) nsupercell
{
  supercell=nsupercell;
  return self;
}

@end


/*
   Implementation for menubar menus, with machinery to optimize updating.
   When an update is initiated, '-mark' is called, which sets a marker in
   menu and all submenus to 0.  As items are parsed from lisp they are
   checked to see if they are the same as existing items.  If so, the
   existing item is kept, otherwise a new replacement item is inserted
   at the marker position.  In either case, the marker is advanced.  Once
   the round is complete, -freeMarked is called, which deletes all items
   ABOVE the marker, which are old items that were not reused.
 */
@implementation EmacsMenu
- init
{
  [super init];
  minmarked=0;
  return self;
}

- (BOOL)performKeyEquivalent:(NSEvent *)theEvent 
{
  if (selected_frame && FRAME_NS_VIEW(selected_frame))
    [FRAME_NS_VIEW(selected_frame) keyDown:theEvent];
  return YES;
}

// inserts item at position 'minmarked' (-addMinItem...)
- addItem: (Lisp_Object) val
   string: (Lisp_Object) aString
     bind: (Lisp_Object)bind
{
  unsigned int key = 0;
  id cell;
  NSString *title, *keyEq;
  EmacsMenuObject *obj;

  if (!NILP(bind))
    {
      Lisp_Object tail,vec;
      for (tail=Fwhere_is_internal(bind,Qnil,Qnil,Qt);
           XTYPE(tail)==Lisp_Cons;
           tail=XCONS(tail)->cdr)
        {
          vec = XCONS(tail)->car;
          if (!VECTORP(vec) ||
              XVECTOR(vec)->size != 1 ||
              XTYPE(XVECTOR(vec)->contents[0]) != Lisp_Int)
            continue;
          key = XFASTINT(XVECTOR(vec)->contents[0]);
          if ((key&(alt_modifier|hyper_modifier|super_modifier))
              ==super_modifier)
            {
              key &= 0xff;
              break;
            }
          else
            {
              key = 0;
            }
        }
    }

  {
      char *muleStr, *utfStr;
      int len;

      muleStr = XSTRING(aString)->data;
      len = strlen(muleStr);
      utfStr = (char *)xmalloc(len*2); // better estimate
      mulemToUTF8s(muleStr, len, utfStr);
      title= [NSString stringWithUTF8String: utfStr];
      xfree(utfStr);
      if (title == nil)
          title = @"<???>";
  }

  keyEq = key == 0 ? @"" : [NSString stringWithFormat:@"%c", key];
//fprintf(stderr, "Adding menuItem '%s' w/keyEq '%s', val is of type %d\n", XSTRING(aString)->data, [keyEq cString], XTYPE(val));
//debug_print(val);
  cell = [self addMinItemWithTitle: title
                            action: @selector(menuDown:)
                     keyEquivalent: keyEq];
  ///[cell setEnabled: YES];

  obj = [[[EmacsMenuObject alloc] init] autorelease];
  [obj setValue:val];
  [obj setSupercell:supercell];
  
  [cell setRepresentedObject:obj];
  
  return cell;
}

- supercell
{
  return supercell;
}

- setSupercell:(id) nsupercell
{
  supercell=nsupercell;
  return self;
}

- mark
{
  NSEnumerator *ienum;
  id <NSMenuItem> item;

  // strictly only needed on GNUstep/Art, but is correct style
  [self setMenuChangedMessagesEnabled: NO];

  ienum = [[self itemArray] objectEnumerator];
  while ( (item = [ienum nextObject]) )
    if ([item hasSubmenu])
      [(EmacsMenu *)[item submenu] mark];
#ifdef MACOSX
  // this protects the app menu
  if ([self supermenu]==nil && ![[self title] isEqualToString: @"Services"])
    minmarked = 1;
  else if ([[self title] length] == 0 ||
           [[self title] isEqualToString: @"Apple"])
    minmarked = [[self itemArray] count];
  else
#endif
    minmarked = 0;

  return self;
}

- freeMarked
{
  int i;
  id cell;
  NSArray *items = [self itemArray];
  int nFreed = [items count] - minmarked;
//if (nFreed > 0) fprintf(stderr, "FreeMarked: %s: freeing %d items\n", [[self title] cString], nFreed);

  i = [items count];
  while (i > minmarked && ((cell=[items objectAtIndex:--i])!=nil))
    [self removeItem:cell];

  for (i=0;i < [items count];i++)
    {
      cell = [items objectAtIndex:i];
      if ([cell hasSubmenu])
        [(EmacsMenu *)[cell submenu] freeMarked];
    }

  // see note above in -mark
  [self setMenuChangedMessagesEnabled: YES];

  //PENDING: track whether anything was added too (seems auto on OS X)
#ifdef MACOSX
  if ([self supermenu] == nil && (nFreed > 0))
    [self sizeToFit];
#else
  if ([[self window] isVisible])
    [self sizeToFit];
#endif
  return self;
}

- findCell:(Lisp_Object)value
{
  NSEnumerator *ienum;
  id <NSMenuItem> item;
  
  ienum = [[self itemArray] objectEnumerator];
  while ( (item = [ienum nextObject]) )
    if ( [(EmacsMenuObject *)[item representedObject] value] == value )
      return item;
  
  return nil;
}

- (int)findCellNumber:(Lisp_Object)value
{
  id cell;
  NSArray *items = [self itemArray];
  int i;
  
  for (i=0; i < [items count]; i++)
    {
      cell = [items objectAtIndex: i];
      if ([(EmacsMenuObject *)[cell representedObject] value] == value) 
        return i;
    }
  return -1;
}

- (id <NSMenuItem>)addMinItemWithTitle: (NSString *)aString 
                                action: (SEL)aSelector 
                         keyEquivalent: (NSString *)charCode
{
  return [self insertItemWithTitle: aString 
                            action: aSelector 
                     keyEquivalent: charCode 
                           atIndex: minmarked];  
}

- removeCell:cell
{
  int i = [self findCellNumber:
                  [(EmacsMenuObject *)[cell representedObject] value]];
  
  if (i != -1)
    {
      [self removeItem: cell];
    
      if (i < minmarked) 
        minmarked--;
    }
  return self;
}

- adjustMap: (Lisp_Object)keymap;
{
  Lisp_Object item;

  for (keymap=XCONS(get_keymap(keymap))->cdr;
       XTYPE(keymap) == Lisp_Cons;
       keymap = XCONS(keymap)->cdr)
    {
      item = XCONS(keymap)->car;

      if (XTYPE(item) == Lisp_Cons)
        {
          [self adjustKey: XCONS(item)->car binding: XCONS(item)->cdr];
        }
      else if (VECTORP(item))
        {
          int len = XVECTOR (item)->size,c;
          for (c=0; c<len; c++)
            [self adjustKey: c binding: XVECTOR(item)->contents[c]];
        }
      // (otherwise item was a string or a separator)
    }
  return self;
}

// called internally, for submenus and items
- (void)adjustKey:(Lisp_Object)key binding:(Lisp_Object)bind
{
  if (XTYPE(bind) == Lisp_Cons)
    [self adjustKey: key binding: XCONS(bind)->cdr string: XCONS(bind)->car];
  else if (EQ (bind, Qundefined))
    [self removeCell:[self findCell:key]];
  return;
}

// called externally, and also internally, to create new menuitems
- (void)adjustKey:(Lisp_Object)key
          binding:(Lisp_Object)bind
           string:(Lisp_Object)string
{
  id cell = nil;
  NSString *str;

  // this path occurs for menus created w/easymenu.el
  if (XTYPE(string) != Lisp_String)
    {
//fprintf(stderr, "Nonstring; symbol name=%d, binding of type %d, car/cdr = %s / %d\n",(XSYMBOL(string)->name)->data, XTYPE(bind), XSTRING(XCONS(bind)->car)->data,XTYPE(XCONS(bind)->cdr));
      if (XTYPE(bind) == Lisp_Cons && !NILP(XCONS(bind)->car))
        {
          if (CONSP(XCONS(bind)->cdr))
            {
              [self adjustKey:key binding:XCONS(XCONS(bind)->cdr)->car
                       string:XCONS(bind)->car];
            }
          else
            {
              [self adjustKey:key binding:XCONS(bind)->cdr
                       string:XCONS(bind)->car];
            }
        }
      return;
    }

  if ([[self itemArray] count] > minmarked
      && ((cell = [[self itemArray] objectAtIndex: minmarked]) != nil) 
      && (([cell isSeparatorItem] &&(!strlen(XSTRING(string)->data) ||
                                     !strncmp("--",XSTRING(string)->data,2)))
          ||
          (([(EmacsMenuObject *)[cell representedObject] value] == key)
           && !strcmp([[cell title] UTF8String], XSTRING(string)->data)
           && [cell isEnabled])))
    {

      /* Keep item, but check (below) that its subitems are the same. */
      if ([cell isSeparatorItem])
        {
          minmarked++;
          return;
        }

    }
  else if (!NILP(Fkeymapp(bind)))
    {
      // new submenu
      cell = [self addItem:key string:string bind:Qnil];
    }
  else
    {
      // not sure what this case is, but xmenu.c also drops it
      if (!(SYMBOLP(bind) || (CONSP(bind) && NILP(Fkeymapp(bind)))))
        return;
    // new single item
      if (!strlen(XSTRING(string)->data) ||
          !strncmp("--",XSTRING(string)->data,2))
        {
          [self insertItem: [NSMenuItem separatorItem] atIndex: minmarked++];
          return;
        }
      else
        {
          cell = [self addItem:key string:string bind:bind];
        }
    }

    {
      char *muleStr, *utfStr;
      int len;

      muleStr = XSTRING(string)->data;
      len = strlen(muleStr);
      utfStr = (char *)xmalloc(len*2); // better estimate
      mulemToUTF8s(muleStr, len, utfStr);
      str = [NSString stringWithUTF8String: utfStr/*XSTRING(string)->data*/];
      xfree(utfStr);
    }

  if (!NILP(Fkeymapp(bind)))
    {
      id nmenu;
     
      if (![cell hasSubmenu])
        {
          nmenu = [[EmacsMenu alloc] initWithTitle: str];
          [nmenu setAutoenablesItems: NO];
          [nmenu setSupercell:cell];         
          [self setSubmenu:nmenu forItem:cell];
          [nmenu release];
        }
      else
        {
          nmenu = [cell submenu];
        }
     
      [nmenu mark];
      [nmenu adjustMap:bind];
    }
  else
    {
      if ([cell hasSubmenu])
        {
          [self removeCell: cell];
          cell = [self addItem:key string:string bind:bind];
        }
    }

  if (![[cell title] isEqualToString: str])
    [cell setTitle:str];

  //PENDING: this is too slow to be done this regularly; menu mgmt logic
  //         will eventually need to be changed to parallel the ugly xmenu.c
//    parse_menu_item(Fcons(string,bind),0,0);
//    if (!NILP(XVECTOR(item_properties)->contents[ITEM_PROPERTY_TYPE]))
//        fprintf(stderr, "Found a toggle item: %s\n", XSTRING(string)->data);
//    [cell setEnabled: 
//             NILP(XVECTOR(item_properties)->contents[ITEM_PROPERTY_ENABLE]) ?
//          NO : YES];
  [cell setEnabled: YES];

  minmarked++;
}

-(void)setSubmenu: (NSMenu *)menu forItem: (id<NSMenuItem>)item
{
  if (![menu isKindOfClass: [EmacsMenu class]])
    {
      // this occurs unavoidably for the services menu; limit the damage
//fprintf(stderr, "not an emacs menu: '%s'\n", [[menu title] cString]);
//[NSException raise: @"NotAnEmacsMenu" format: @"Not an EmacsMenu. "];
      [menu setAutoenablesItems: NO];
      [menu retain];
    }
  else
    {
      [super setSubmenu: menu forItem: item];
    }
}

- dump: (NSString *)indent
{
  NSEnumerator *ienum;
  id <NSMenuItem> item;
  NSString *subindent = [indent stringByAppendingString: @"    "];
  int pos = 0;

  ienum = [[self itemArray] objectEnumerator];
  while ( (item = [ienum nextObject]) )
    {
      BOOL hasSub = [item hasSubmenu];
      if (pos++ == minmarked)
        fprintf(stderr, "%sMARK\n", [indent cString]);
      fprintf(stderr, "%s%s%s\n",[indent cString],[[item title] UTF8String],
              (hasSub ? " >" : ""));
      if (hasSub)
        {
          NSMenu *sub = [item submenu];
          if (![sub isKindOfClass: [EmacsMenu class]])
            fprintf(stderr, "%s(not an EmacsMenu)\n", [subindent cString]);
          else
            [(EmacsMenu *)sub dump: subindent];
        }
    }
  return self;
}

@end


@implementation EmacsMenuCard
- init
{
  [super init];
  [self setFont:[NSFont userFixedPitchFontOfSize:0]];
  [self setTitle:@""];
  [self setPullsDown:YES];
  [self setTarget:self];
  [self setAction:@selector(clicked:)];
  [self setAlignment:NSCenterTextAlignment];
#ifdef MACOSX
  //PENDING: can now select from whole menu, but don't see "Select"
  [self addItem: "Select..." value: 0 enabled: YES];
#endif
  return self;
}

- addTitle: (char *)name
{
  char *utfStr;
  int len;
  NSString *title;

  len = strlen(name);
  utfStr = (char *)xmalloc(len*2); // better estimate
  mulemToUTF8s(name, len, utfStr);
  title= [NSString stringWithUTF8String: utfStr];
  xfree(utfStr);
  if (title == nil)
    title = @"<???>";

  [self setTitle: title];
  return self;
}

- addItem: (char *)name value:(Lisp_Object)val enabled:(BOOL)enabled
{
  NSRect r;
  id item;
  float f;
  char *utfStr;
  int len;
  NSString *title;

  len = strlen(name);
  utfStr = (char *)xmalloc(len*2); // better estimate
  mulemToUTF8s(name, len, utfStr);
  title= [NSString stringWithUTF8String: utfStr];
  xfree(utfStr);
  if (title == nil)
      title = @"<???>";

  [self addItemWithTitle: title];
  item = [self itemAtIndex:[self numberOfItems]-1];
  [item setTag:(int)val];

  if (!enabled)
      [item setEnabled:NO];

  //PENDING: make sure enough room for title (+ buttons) too
  f = [[self font] widthOfString: [item title]] + 12.0;
  r = [self frame];
  if (f > r.size.width)
    [self setFrameSize: NSMakeSize(f, r.size.height)];

  return self;
}

- (void)clicked:sender
{
  Lisp_Object seltag;
  id selected;
  
  if ( (selected = [self selectedItem]) && [selected tag] )
    {
      seltag = (Lisp_Object)[selected tag];
      if (! EQ(seltag,Qundefined))
        [NSApp stopModalWithCode:seltag];
    }
}
@end

@implementation EmacsMenuPanel
- initWithContentRect: (NSRect)contentRect 
            styleMask: (unsigned int)aStyle 
              backing: (NSBackingStoreType)backingType 
                defer: (BOOL)flag
{
  aStyle=NSTitledWindowMask;
  flag=YES;
  list=nil;
  maxwidth=0;
  [super initWithContentRect:contentRect 
                   styleMask:aStyle|NSClosableWindowMask 
                     backing:backingType 
                       defer:flag];
  [self setOneShot:YES];
  [self setReleasedWhenClosed:YES];
  [self setHidesOnDeactivate:YES];
  return self;
}

- (BOOL)windowShouldClose:(id)sender
{
  [NSApp stopModalWithCode:Qnil];
  return NO;
}

static void process_keymap(id window,Lisp_Object keymap,Lisp_Object prefix,
                           char *name)
{
  Lisp_Object tail,pending_maps,tem;
  int card;
  
  pending_maps=Qnil;
  
  card = [window cards];
  if (name)
    [window addTitle:name to:card];
  
  for (tail = get_keymap(keymap); XTYPE(tail) == Lisp_Cons;
       tail = XCONS(tail)->cdr)
    {
      Lisp_Object item = XCONS(tail)->car, item1, item2, def, enabled;
      if (XTYPE(item) == Lisp_Cons)
        {
          item1 = XCONS(item)->cdr;
          if (XTYPE(item1) == Lisp_Cons)
            {
              item2 = XCONS(item1)->car;
              if (XTYPE(item2) == Lisp_String)
                {
                  def = XCONS(item1)->cdr;
                  enabled = Qt;
                  if (XTYPE(def) == Lisp_Symbol)
                    {
                      tem = Fget(def, Qmenu_enable);
                      if (!NILP(tem))
                        enabled = Feval(tem);
                    }
                  if (XSTRING (item2)->data[0] == '@' && !NILP(Fkeymapp(def)))
                    pending_maps = Fcons(Fcons(def,
                                               Fcons(item2, XCONS(item)->car)),
                                         pending_maps);
                  else
                    [window addItem: XSTRING(item2)->data
                              value: Freverse(Fcons(XCONS(item)->car, prefix))
                            enabled: !NILP(enabled)
                                 at:card];
                }
            }
        }
      else if (VECTORP(item))
        {
          int len = XVECTOR(item)->size;
          int c;
          for (c = 0; c < len; c++)
            {
              Lisp_Object character;
              XSETFASTINT(character, c);
              item1 = XVECTOR(item)->contents[c];
              if (XTYPE(item1) == Lisp_Cons)
                {
                  item2 = XCONS(item1)->car;
                  if (XTYPE(item2) == Lisp_String)
                    {
                      def = XCONS(item1)->cdr;
                      enabled = Qt;
                      if (XTYPE(def) == Lisp_Symbol)
                        {
                          tem = Fget(def,Qmenu_enable);
                          if (!NILP(tem))
                            enabled = Feval(tem);
                        }
                      if (XSTRING (item2)->data[0] == '@' &&
                          !NILP(Fkeymapp(def)))
                        pending_maps = Fcons(Fcons(def,
                                                   Fcons(item2,
                                                         XCONS(item)->car)),
                                             pending_maps);
                      else
                        [window addItem: XSTRING(item2)->data
                                  value: Freverse(Fcons(XCONS(item)->car,prefix))
                                enabled: !NILP(enabled)
                                     at:card];
                    }
                }
            }
        }
    }
  
  for (tail = pending_maps; XTYPE(tail) == Lisp_Cons; tail = XCONS(tail)->cdr)
    {
      tem = XCONS(tail)->car;
      process_keymap(window,
                     XCONS(tem)->car,
                     Fcons(XCONS(XCONS(tem)->cdr)->cdr,prefix),
                     (char *)XSTRING(XCONS(XCONS(tem)->cdr)->car)->data + 1);
    }
}

void process_list(id window,Lisp_Object list)
{
  Lisp_Object tail,pane,tem;
  int card;
  
  card = [window cards];
  for ( ; XTYPE(list) == Lisp_Cons; list = XCONS(list)->cdr)
    {
      pane = XCONS(list)->car;
      if (XTYPE(pane) != Lisp_Cons || XTYPE(XCONS(pane)->car) != Lisp_String)
        continue;
      for (tail = pane; XTYPE(tail) == Lisp_Cons; tail = XCONS(tail)->cdr)
        {
          tem = XCONS(tail)->car;
          if (XTYPE(tem) == Lisp_String)
            {
              if (tail != pane)		/* Not very first item? */
                [window addItem: XSTRING(tem)->data
                          value: Qnil
                        enabled: NO
                             at: card];
              else
                [window addTitle: XSTRING(tem)->data to: card];
            }
          else if (XTYPE(tem) == Lisp_Cons ||
                   XTYPE(XCONS(tem)->car) == Lisp_String)
            {
              [window addItem: XSTRING(XCONS(tem)->car)->data
                        value: XCONS(tem)->cdr
                      enabled:YES
                           at: card];
            }
        }
    }
}


- addItem:(char *)str value:(Lisp_Object)val enabled:(BOOL)enabled at:(int)pane
{
  int row;
  EmacsMenuCard *card;
  NSRect area;
  
  for (row = [list count]; row <= pane; row++)
    {
      card = [[EmacsMenuCard alloc] init];
      [list addObject: card];
      [[self contentView] addSubview:card];

      area.size = [[card cell] cellSize];
      if (area.size.width > maxwidth)
        maxwidth=area.size.width;
      area.origin.x = MenuStagger*pane;
      area.origin.y = area.size.height*pane;
      [card setFrame:area];
    }

  card = [list objectAtIndex:pane];
  [card addItem:str value:val enabled:enabled];
  area = [card frame];
  if (area.size.width > maxwidth)
    maxwidth=area.size.width;

  return self;
}

- addTitle:(char *)str to:(int)pane
{
  int row;
  EmacsMenuCard *card;
  NSRect area;
  
  for (row = [list count]; row <= pane; row++)
    {
      card=[[EmacsMenuCard alloc] init];
      [list addObject: card];
      [[self contentView] addSubview:card];

      area.size = [[card cell] cellSize];
      if (area.size.width > maxwidth)  maxwidth=area.size.width;
      area.origin.x = MenuStagger*pane;
      area.origin.y = area.size.height*pane;
      [card setFrame:area];
    }
  
  card = [list objectAtIndex:pane];
  [card addTitle:str];

  area.size = [[card cell] cellSize];
  if (area.size.width > maxwidth) 
    maxwidth=area.size.width;

  return self;
}

- initFromMenu:(Lisp_Object)menu
{
  Lisp_Object title;
  [super init];

  list = [[NSMutableArray alloc] init];
  
  if (!NILP (Fkeymapp(menu)))
    {
      Lisp_Object keymap = get_keymap(menu);
      process_keymap(self,keymap,Qnil,0);
      title = map_prompt(keymap);
    }
  else if (XTYPE(menu)==Lisp_Cons && !NILP(Fkeymapp (Fcar(menu))))
    {
      Lisp_Object tem,keymap;
      for (title=Qnil;XTYPE(menu) == Lisp_Cons;menu = Fcdr(menu))
        {
          keymap=get_keymap(Fcar(menu));
          process_keymap(self,keymap,Qnil,0);
          tem = map_prompt (keymap);
          if (NILP(title) && !NILP(tem)) title=tem;
        }
    }
  else
    {
      title = Fcar(menu);
      process_list(self,Fcdr(menu));
    }
  
  if (XTYPE(title)==Lisp_String)
    {
      char *muleStr, *utfStr;
      int len;
      NSString *titleStr;

      muleStr = XSTRING(title)->data;
      len = strlen(muleStr);
      utfStr = (char *)xmalloc(len*2); // better estimate
      mulemToUTF8s(muleStr, len, utfStr);
      titleStr = [NSString stringWithUTF8String: utfStr];
      xfree(utfStr);
      if (titleStr == nil)
        titleStr = @"<???>";

      [self setTitle: titleStr];
    }
  
  {
    int i;
    NSRect r,s,t;
    
    if ([list count] > 0)
      {
        t = [(EmacsMenuCard *)[list objectAtIndex:0] frame];
        for (i = 0; i < [list count]; i++)
          [[list objectAtIndex:i] 
                  setFrameSize:NSMakeSize(maxwidth, t.size.height)];
      }
    else
      {
        t = NSMakeRect(0,0,0,0);
        fprintf(stderr, "WARNING: menu panel w/no title\n");
      }
    r = [self frame];
    s = [(NSView *)[self contentView] frame];
    r.size.height += t.size.height*i-s.size.height;
    r.size.width  += maxwidth+(i-1)*MenuStagger-s.size.width;
    [self setFrame:r display:YES];
  }
  
  return self;
}

- (void)dealloc
{
  [list release];
  [super dealloc]; 
  return;
}

- (int)cards
{
  return [list count];
}

- (Lisp_Object)runMenuAt:(float)x:(float)y
{
  NSEvent *e;
  NSModalSession session;
  NSRect r;
  int ret;
  
  if ([list count] == 0)
    return Qnil;
  
  r = [(EmacsMenuCard *)[list objectAtIndex:0] frame];
  r.origin = [[self contentView] convertPoint:r.origin toView:nil];
  
  [self setFrameOrigin:NSMakePoint(x-r.size.width/2.0,
                                   y-r.origin.y+r.size.height/2.0)];
  [self orderFront:NSApp];
  
  session = [NSApp beginModalSessionForWindow:self];

  [self setHidesOnDeactivate:YES];
  while ((ret=[NSApp runModalSession:session])==NSRunContinuesResponse)
    (e = [NSApp nextEventMatchingMask:NSAnyEventMask
                            untilDate:[NSDate distantFuture]
                               inMode:NSEventTrackingRunLoopMode
                              dequeue:(!1)]);
  
  [NSApp endModalSession:session];

  return (Lisp_Object)ret;
}
@end

@implementation EmacsDialogPanel

#define SPACER		8.0
#define ICONSIZE	50.0
#define TEXTHEIGHT	20.0

- initWithContentRect:(NSRect)contentRect styleMask:(unsigned int)aStyle
              backing:(NSBackingStoreType)backingType defer:(BOOL)flag
{
  NSSize spacing = {SPACER,SPACER};
  NSRect area;
  char this_cmd_name[80];
  id cell,tem;
  static NSImageView *imgView;

  if (imgView == nil)
    {
      NSImage *img;
      area.origin.x   = SPACER;
      area.origin.y   = SPACER;
      area.size.width = ICONSIZE;
      area.size.height= ICONSIZE;
      img = [[NSImage imageNamed: @"NSApplicationIcon"] copy];
      [img setScalesWhenResized: YES];
      [img setSize: NSMakeSize(ICONSIZE, ICONSIZE)];
      imgView = [[NSImageView alloc] initWithFrame: area];
      [imgView setImage: img];
      [imgView setEditable: NO];
      [img release];
    }

  aStyle=NSTitledWindowMask;
  flag=YES;
  rows=0;
  cols=1;
  [super initWithContentRect:contentRect styleMask:aStyle
                     backing:backingType defer:flag];
  [[self contentView] setAutoresizesSubviews:YES];

  [[self contentView] addSubview: imgView];
  [self setTitle: @"Emacs"];

  if (XTYPE(Vthis_command)==Lisp_Symbol)
    {
      int i;

      strcpy(this_cmd_name,XSYMBOL(Vthis_command)->name->data);
      if (this_cmd_name[0] >= 'a' && this_cmd_name[0] <= 'z')
        this_cmd_name[0] = this_cmd_name[0]-'a'+'A';
      for (i=0; this_cmd_name[i] != '\0'; i++)
        {
          if (this_cmd_name[i] == '-')
            {
              this_cmd_name[i]=' ';
              if (this_cmd_name[i+1] >= 'a' && this_cmd_name[i+1] <= 'z')
                this_cmd_name[i+1] = this_cmd_name[i+1]-'a'+'A';
            }
        }
    }
  else
    strcpy(this_cmd_name,"Emacs");
  area.origin.x   = ICONSIZE+2*SPACER;
  area.origin.y   = ICONSIZE/2-10+SPACER;
  area.size.width = 400;
  area.size.height= TEXTHEIGHT;
  command = [[NSTextField alloc] initWithFrame:area];
  [[self contentView] addSubview:command];
  [command setStringValue:[NSString stringWithUTF8String:this_cmd_name]];
  [command setDrawsBackground:NO];
  [command setBezeled:NO];
  [command setSelectable:NO];
  [command setFont:[NSFont systemFontOfSize:18.0]];

  area.origin.x   = 0;
  area.origin.y   = ICONSIZE+2*SPACER;
  area.size.width = 400;
  area.size.height= 2;
  tem = [[NSBox alloc] initWithFrame:area];
  [[self contentView] addSubview:tem];
  [tem setTitlePosition:NSNoTitle];
  [tem setAutoresizingMask:NSViewWidthSizable];

  area.origin.x   = 2*SPACER;
  area.origin.y  += 2*SPACER;
  area.size.width = 400;
  area.size.height= TEXTHEIGHT;
  title = [[NSTextField alloc] initWithFrame:area];
  [[self contentView] addSubview:title];
  [title setDrawsBackground:NO];
  [title setBezeled:NO];
  [title setSelectable:NO];
  [title setFont:[NSFont systemFontOfSize:14.0]];

  cell = [[NSButtonCell alloc] initTextCell:@""];
  // Make it look like nothing's here
  [cell setBordered:NO];
  [cell setEnabled:NO];
  [cell setCellAttribute:NSCellIsInsetButton to:8];

  matrix = [[NSMatrix alloc] initWithFrame:contentRect 
                                      mode:NSHighlightModeMatrix 
                                 prototype:cell 
                              numberOfRows:0 
                           numberOfColumns:1];
  [[self contentView] addSubview:matrix];
  [matrix setFrameOrigin:NSMakePoint(SPACER, area.origin.y+TEXTHEIGHT+2*SPACER)];
  [matrix setIntercellSpacing:spacing];

  [self setOneShot:YES];
  [self setReleasedWhenClosed:YES];
  [self setHidesOnDeactivate:YES];
  return self;
}

- (BOOL)windowShouldClose:(id)sender
{
  [NSApp stopModalWithCode:Qnil];
  return NO;
}

void process_dialog(id window,Lisp_Object list)
{
  Lisp_Object item;
  int row;

  row=0;
  for (; XTYPE(list)==Lisp_Cons; list=XCONS(list)->cdr)
    {
      item=XCONS(list)->car;
      if (XTYPE(item)==Lisp_String)
        {
          [window addString:XSTRING(item)->data row:row];
          row++;
        }
      else if (XTYPE(item)==Lisp_Cons)
        {
          [window addButton:XSTRING(XCONS(item)->car)->data
                      value:XCONS(item)->cdr row:row];
          row++;
        }
      else if (NILP(item))
        {
          [window addSplit];
          row=0;
        }
    }
}


- addButton:(char *)str value:(Lisp_Object)val row:(int)row
{
  id cell;
       
  if (row >= rows)
    {
      [matrix addRow];
      rows++;
    }
  cell = [matrix cellAtRow:row column:cols-1];
  [cell setTarget:self];
  [cell setAction:@selector(clicked:)];
  [cell setTitle:[NSString stringWithUTF8String:str]];
  [cell setTag:(int)val];
  [cell setBordered:YES];
  [cell setEnabled:YES];

  return self;
}

- addString:(char *)str row:(int)row
{
  id cell;
       
  if (row >= rows)
    {
      [matrix addRow];
      rows++;
    }
  cell = [matrix cellAtRow:row column:cols-1];
  // No setAction here.  Disabled buttons can't trigger actions
  [cell setTitle:[NSString stringWithUTF8String:str]];
  [cell setBordered:YES];
  [cell setEnabled:NO];

  return self;
}

- addSplit
{
  [matrix addColumn];
  cols++;
  return self;
}

- clicked:sender
{
  NSArray *sellist=nil;
  Lisp_Object seltag;

  sellist=[sender selectedCells];
  if ([sellist count]<1) 
    return self;

  seltag = (Lisp_Object)[[sellist objectAtIndex:0] tag];
  if (! EQ(seltag,Qundefined))
    [NSApp stopModalWithCode:seltag];
  return self;
}

- initFromContents:(Lisp_Object)contents
{
  Lisp_Object head;
  [super init];

  if (XTYPE(contents)==Lisp_Cons)
    {
      head = Fcar(contents);
      process_dialog(self,Fcdr(contents));
    }
  else
    head = contents;

  if (XTYPE(head)==Lisp_String)
    [title setStringValue:
             [NSString stringWithUTF8String:XSTRING(head)->data]];

  {
    int i;
    NSRect r,s,t;

    if (cols == 1 && rows > 1)	// Never told where to split
      {
        [matrix addColumn];
        for (i=0; i<rows/2; i++)
          {
            [matrix putCell:[matrix cellAtRow:(rows+1)/2 column:0]
                      atRow:i column:1];
            [matrix removeRow:(rows+1)/2];
          }
      }
    [matrix sizeToFit];
    [title sizeToFit];
    [command sizeToFit];

    t = [matrix frame];
    r = [title frame];
    if (r.size.width+r.origin.x > t.size.width+t.origin.x)
      {
        t.origin.x   = r.origin.x;
        t.size.width = r.size.width;
      }
    r = [command frame];
    if (r.size.width+r.origin.x > t.size.width+t.origin.x)
      {
        t.origin.x   = r.origin.x;
        t.size.width = r.size.width;
      }

    r = [self frame];
    s = [(NSView *)[self contentView] frame];
    r.size.height += t.origin.y+t.size.height+SPACER-s.size.height;
    r.size.width  += t.origin.x+t.size.width +SPACER-s.size.width;
    [self setFrame:r display:NO];
  }

  return self;
}

- (void)dealloc
{
  [matrix release];
  { [super dealloc]; return; };
}

- (Lisp_Object)runDialogAt:(float)x:(float)y
{
  NSEvent *e;
  NSModalSession session;
  int ret;

  [self center];
  [self orderFront:NSApp];

  session = [NSApp beginModalSessionForWindow:self];
  while ((ret=[NSApp runModalSession:session])==NSRunContinuesResponse)
    (e = [NSApp nextEventMatchingMask:NSAnyEventMask
                            untilDate:[NSDate distantFuture]
                               inMode:NSEventTrackingRunLoopMode
                              dequeue:NO]);
  [NSApp endModalSession:session];

  return (Lisp_Object)ret;
}

@end
